/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CPU_H_
#define _CTRL_CPU_H_

#include <omp.h>
#include <semaphore.h>
#include <pthread.h>

#include "hitmap.h"

#include "Core/Ctrl_Policy.h"
#include "Core/Ctrl_TaskQueue.h"

#define CTRL_CPU_LAUNCH( p_ctrl, name, threads, ... ) \
	case CTRL_TYPE_CPU: \
		Ctrl_LaunchKernel(p_ctrl, Ctrl_KernelTaskCreate_##name( CTRL_TYPE_CPU, threads, CTRL_KERNEL_ARGS_TO_POINTERS( __VA_ARGS__) )); \
		break;

typedef struct {
	int								n_cores;
	cpu_set_t 						cpuset;

	int								n_task;
                          /**< Semáforo para control de ejecución de hilo principal.*/
	sem_t 							sem_sync;
	sem_t							sem_destroy;

	omp_lock_t						lock_queue;

	Ctrl_Policy						policy;
	
	Ctrl_TaskQueue					task_queue;
} Ctrl_Cpu;


#ifdef __cplusplus
	extern "C" {
#endif

void Ctrl_Cpu_Create(Ctrl_Cpu *p_ctrl, int n_cores, Ctrl_Policy policy);

void Ctrl_Cpu_LaunchKernel(Ctrl_Cpu *p_ctrl, Ctrl_Task task);

void Ctrl_Cpu_Destroy(Ctrl_Cpu *p_ctrl);


void Ctrl_Cpu_GlobalSync(Ctrl_Cpu *p_ctrl);	


void Ctrl_Cpu_Attach(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_MoveTo(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_Detach(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_MoveFrom(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_CreateInternal(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_DestroyInternal(Ctrl_Cpu *p_ctrl, HitTile *p_tile);


void Ctrl_Cpu_WaitTile(Ctrl_Cpu *p_ctrl, HitTile *p_tile);	//TODO @waxa from @ismael, unused	

#ifdef __cplusplus
	}
#endif

#endif /* _CTRL_CPU_H_ */
