/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CUDA_H_
#define _CTRL_CUDA_H_

#include <omp.h>
#include <stdbool.h>

#ifdef _CTRL_CUDA_DEBUG_
    #include <stdio.h>
#endif

#include <cuda_runtime_api.h>

#include "hitmap2.h"

#include "Core/Ctrl_Policy.h"
#include "Core/Ctrl_Type.h"
#include "Core/Ctrl_TaskQueue.h"
#include "Core/Ctrl_Tile.h"
#include "Core/Ctrl_KHitTile.h"
#include "Core/Ctrl_Request.h"

#include "Kernel/Ctrl_KernelProto.h"
#include "Kernel/Ctrl_ImplType.h"

#include "Architectures/Cuda/Ctrl_Cuda_Request.h"
#include "Architectures/Cuda/Ctrl_Cuda_Helper.h"
#include "Architectures/Cuda/Ctrl_Cuda_Tile.h"


#define CTRL_CUDA_LAUNCH(p_ctrl, name, threads, block, ... ) \
    case CTRL_TYPE_CUDA: \
        CTRL_KERNEL_CUDA_CHAR_block_##name[0].x = block.x; \
        CTRL_KERNEL_CUDA_CHAR_block_##name[0].y = block.y; \
        CTRL_KERNEL_CUDA_CHAR_block_##name[0].z = block.z; \
        Ctrl_LaunchKernel(p_ctrl, Ctrl_KernelTaskCreate_##name( CTRL_TYPE_CUDA, threads, CTRL_KERNEL_ARGS_TO_POINTERS( __VA_ARGS__) )); \
        break; 


typedef struct Ctrl_Cuda {
    int 							device;

    cudaStream_t 					stream_kernel;
    cudaStream_t 					stream_host;

	cudaEvent_t						event_seq;

    struct Ctrl_Cuda_Tile_List		*p_tile_list_head;
	struct Ctrl_Cuda_Tile_List		*p_tile_list_tail;

	omp_lock_t						*p_lock_first_host;
	omp_lock_t						*p_lock_first_ctrl;
	omp_lock_t						*p_lock_host;
	omp_lock_t						*p_lock_ctrl;
	
    Ctrl_Policy						policy;
} Ctrl_Cuda;

void Ctrl_Cuda_Create(Ctrl_Cuda *p_ctrl, Ctrl_Policy policy, int device);
void Ctrl_Cuda_Destroy(Ctrl_Cuda *p_ctrl);
void Ctrl_Cuda_EvalTask(Ctrl_Cuda *p_ctrl, Ctrl_Task *p_task);
void Ctrl_Cuda_NewThreadSetup(Ctrl_Cuda *p_ctrl);

#endif /* _CTRL_CUDA_H_ */
