/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_OPENCL_GPU_H_
#define _CTRL_OPENCL_GPU_H_

#include <omp.h>
#include <stdbool.h>

#ifdef _CTRL_OPENCL_GPU_DEBUG_
	#include <stdio.h>
#endif

#ifndef CL_USE_DEPRECATED_OPENCL_1_2_APIS
#define CL_USE_DEPRECATED_OPENCL_1_2_APIS
#endif // CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>

#include "hitmap2.h"

#include "Core/Ctrl_Policy.h"
#include "Core/Ctrl_Type.h"
#include "Core/Ctrl_TaskQueue.h"
#include "Core/Ctrl_Tile.h"
#include "Core/Ctrl_KHitTile.h"
#include "Core/Ctrl_Request.h"

#include "Kernel/Ctrl_KernelProto.h"
#include "Kernel/Ctrl_ImplType.h"

#include "Architectures/OpenCL/Ctrl_OpenCL_Request.h"
#include "Architectures/OpenCL/Ctrl_OpenCL_Helper.h"
#include "Architectures/OpenCL/Ctrl_OpenCL_Tile.h"


#define CTRL_OPENCL_GPU_LAUNCH( p_ctrl, name, threads, group, ... ) \
	case CTRL_TYPE_OPENCL_GPU: \
		local_size_OPENCL_GPU_##name[0] = group.x; \
		local_size_OPENCL_GPU_##name[1] = group.y; \
		local_size_OPENCL_GPU_##name[2] = group.z; \
		Ctrl_LaunchKernel(p_ctrl, Ctrl_KernelTaskCreate_##name( CTRL_TYPE_OPENCL_GPU, threads, CTRL_KERNEL_ARGS_TO_POINTERS( __VA_ARGS__) )); \
		break;


/*
 * @brief Contiene el estado de un comunicador específico de GPU.
 */
typedef struct Ctrl_OpenCLGpu {
	cl_platform_id						platform_id;
	cl_device_id						device_id;

	cl_context							context;

	cl_command_queue_properties			queue_properties;
	cl_command_queue					queue;

	struct Ctrl_OpenCL_Tile_List		*p_tile_list_head;
	struct Ctrl_OpenCL_Tile_List		*p_tile_list_tail;

	cl_event							last_kernel_event;
	cl_event							last_host_task_event;
	cl_event							default_event;

    omp_lock_t							*p_lock_first_host;
	omp_lock_t							*p_lock_first_ctrl;
	omp_lock_t							*p_lock_host;
	omp_lock_t							*p_lock_ctrl;
	
	Ctrl_Policy							policy;

	Ctrl_OpenCL_Res						*p_resources_list;

	#ifdef _CTRL_OPENCL_GPU_PROFILING_
		int							platform;
		int							device;

		cl_profiling_info 			profiling_info_start;
		cl_profiling_info 			profiling_info_end;

		bool 						has_first_profiling_event;

		cl_event 					first_profiling_event;
		cl_event 					last_profiling_event;

		cl_event 					*profiling_read_events;
		cl_event 					*profiling_write_events;
		cl_event 					*profiling_kernel_events;

		int 						i_read_task;
		int 						i_write_task;
		int 						i_kernel_task;

		cl_ulong 					profiling_total;
		cl_ulong 					profiling_sum;
		cl_ulong 					profiling_offloading;
		cl_ulong 					profiling_read;
		cl_ulong 					profiling_write;
		cl_ulong 					profiling_kernel;

		cl_ulong 					profiling_start;
		cl_ulong 					profiling_end;
		
		#ifdef _CTRL_OPENCL_GPU_PROFILING_VERBOSE_
			visual_event 			*profiling_visual_events;
			int 					i_visual_task;
		#endif
	#endif // _CTRL_EXAMPLES_OPENCL_GPU_PROFILING_
} Ctrl_OpenCLGpu;

#ifdef __cplusplus
	extern "C" {
#endif

void Ctrl_OpenCLGpu_Create(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Policy policy, int device, int platform);
void Ctrl_OpenCLGpu_Destroy(Ctrl_OpenCLGpu *p_ctrl);
void Ctrl_OpenCLGpu_EvalTask(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_InitHostTaskThread(Ctrl_OpenCLGpu *p_ctrl);

#ifdef __cplusplus
	}
#endif

#endif /* _CTRL_OPENCL_GPU_H_ */
