/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_XPHI_H_
#define _CTRL_XPHI_H_

#include <semaphore.h>
#include <omp.h>

#include <offload.h>

#include "hitmap.h"

#include "Core/Ctrl_Policy.h"
#include "Core/Ctrl_TaskQueue.h"

// #define ALLOC alloc_if(1)
// #define REUSE alloc_if(0)
// #define RETAIN free_if(0)
// #define FREE free_if(1)

#define CTRL_XPHI_LAUNCH( p_ctrl, name, threads, ... ) \
		case CTRL_TYPE_XPHI: \
			Ctrl_LaunchKernel(p_ctrl, Ctrl_KernelTaskCreate_##name( CTRL_TYPE_XPHI, threads, CTRL_KERNEL_ARGS_TO_POINTERS( __VA_ARGS__) )); \
			break;

/*
 * @brief Contiene el estado de un comunicador específico de GPU.
 */
typedef struct {

	_Offload_stream					stream;

	int						n_xphi;			//Coprocessor card no. to which this controller is linked
	int						n_signal;		//
	int						default_signal;		//Number tied to a signal already executed
	//uint32_t					n_engines;
	
	int						n_task;
                          /**< Semáforo para control de ejecución de hilo principal.*/
	sem_t						sem_sync;
	sem_t						sem_destroy;

	omp_lock_t					lock_queue;
	
	Ctrl_TaskQueue					task_queue;

	Ctrl_Policy					policy;
} Ctrl_XPhi;

#ifdef __cplusplus
	extern "C" {
#endif
/*
 * @brief Crea un comunicador de GPU.
 * @param comm Comunicador que se va a crear.
 */
void Ctrl_XPhi_Create(Ctrl_XPhi *p_ctrl, int n_xphi, Ctrl_Policy policy);


void Ctrl_XPhi_LaunchKernel(Ctrl_XPhi *p_ctrl, Ctrl_Task task);


void Ctrl_XPhi_Destroy(Ctrl_XPhi *p_ctrl);


void Ctrl_XPhi_GlobalSync(Ctrl_XPhi *p_ctrl);	


/*
 * @brief Recupera stream de CUDA asignado a comunicaciones de un hitTile.
 * @param comm Comunicador utilizado en la comunicaciones. (Uso temporal como parámetro)
 * @param tile HitTile del que se quiere recuperar el stream.
 * @param pstream Puntero a stream donde se va a almacenar.
 */
// void Ctrl_XPhi_GetTileHandler(Ctrl_XPhi *p_ctrl, HitTile *p_tile);
			

/*
 * @brief Destruye el stream de CUDA asignado a comunicaciones de un hitTile.
 * @param comm Comunicador utilizado en la comunicaciones. (Uso temporal como parámetro)
 * @param tile HitTile del que se quiere recuperar el stream.
 */
// void Ctrl_XPhi_DestroyTileHandler(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


void Ctrl_XPhi_Attach(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


void Ctrl_XPhi_AllocPinned(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


void Ctrl_XPhi_FreePinned(Ctrl_XPhi *p_ctrl, HitTile *p_tile);
// Created by ismael @ 2017-12-15 14:13:13
// Mask stream's use
//Yuri
// void Ctrl_XPhi_AttachAsync(Ctrl_XPhi *p_ctrl, HitTile *p_tile); //TODO @waxa remove candidate


//Yuri
void Ctrl_XPhi_MoveTo(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


// Created by ismael @ 2017-12-15 14:13:37
// Mask stream's use
// Yuri
// void Ctrl_XPhi_MoveToAsync(Ctrl_XPhi *p_ctrl, HitTile *p_tile); //TODO @waxa remove candidate


void Ctrl_XPhi_Detach(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


// Created by ismael @ 2017-12-15 14:13:55
// Mask stream's use
// Yuri
// void Ctrl_XPhi_DetachAsync(Ctrl_XPhi *p_ctrl, HitTile *p_tile); //TODO @waxa remove candidate


// Yuri
void Ctrl_XPhi_MoveFrom(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


// Created by ismael @ 2017-12-15 14:14:11
// Mask stream's use
// Yuri
// void Ctrl_XPhi_MoveFromAsync(Ctrl_XPhi *p_ctrl, HitTile *p_tile); //TODO @waxa remove candidate


// Created by ismael @ 2017-12-15 14:14:29
// Implement async version
void Ctrl_XPhi_CreateInternal(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


// Created by ismael @ 2017-12-15 14:14:48
// Implement async version
void Ctrl_XPhi_DestroyInternal(Ctrl_XPhi *p_ctrl, HitTile *p_tile);


//Yuri
void Ctrl_XPhi_WaitTile(Ctrl_XPhi *p_ctrl, HitTile *p_tile);	//TODO @waxa from @ismael, unused	

#ifdef __cplusplus
	}
#endif

#endif // _CTRL_XPHI_H_
