/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CORE_KHITTILE_H_
#define _CTRL_CORE_KHITTILE_H_


// typedef struct KHitTile {
// 	void	* data;
// 	int	origAcumCard[4];
// 	int 	card[3];	
// } KHitTile;

#define CTRL_FPGA_STRUCT_ALIGNMENT 16

#ifndef CTRL_FPGA_KERNEL_FILE
typedef struct {
	void	* data;
	int	origAcumCard[4];
	int 	card[3];	
} KHitTile;
#endif

// #define hit_ktileNewType( type ) \
// typedef struct KHitTile_##type { \
// 	type	* data; \
// 	int	origAcumCard[4]; \
// 	int 	card[3]; \
// } KHitTile_##type;\
// char *raw_ktile_KHitTile_##type = CTRL_KERNEL_STRINGIFY( type ); \
// char *raw_def_ktile_KHitTile_##type = CTRL_KERNEL_STRINGIFY( \
// 		typedef struct KHitTile_##type { \
// 			type	* data; \
// 			int	origAcumCard[4]; \
// 			int 	card[3]; \
// 		} KHitTile_##type; \
// 	); \
// bool raw_added_ktile_KHitTile_##type = false;


/* @author: Gabriel Rodriguez-Canal
   @brief: KHitTile wrapper to isolate tile coordinates from data, as structures with pointers cannot be passed as arguments */
#ifdef _CTRL_ARCH_FPGA_
	typedef struct __attribute__((packed)) __attribute__((aligned(CTRL_FPGA_STRUCT_ALIGNMENT))) {
		int	origAcumCard[4];
		int 	card[3];
	} KHitTile_fpga_wrapper;
#endif
#ifdef _CTRL_ARCH_OPENCL_GPU_
	typedef struct {
		int	origAcumCard[4];
		int 	card[3];
	} KHitTile_opencl_wrapper;
#endif


/* @author: Gabriel Rodriguez-Canal
   @brief: FPGA architectures cannot deal with strings, as they execute binaries. Strings are turned into actual C code */
#ifndef _CTRL_ARCH_FPGA_
	#define hit_ktileNewType( type ) \
		typedef struct { \
			type	* data; \
			int	origAcumCard[4]; \
			int 	card[3]; \
		} KHitTile_##type;\
		char *raw_ktile_KHitTile_##type = CTRL_KERNEL_STRINGIFY( type ); \
		char *raw_def_ktile_KHitTile_##type = CTRL_KERNEL_STRINGIFY( \
				typedef struct { \
					__global type	* data; \
					int	origAcumCard[4]; \
					int 	card[3]; \
				} KHitTile_##type##_write; \
				typedef struct { \
					__global const type	* data; \
					int	origAcumCard[4]; \
					int 	card[3]; \
				} KHitTile_##type##_read; \
				typedef struct { \
					int	origAcumCard[4]; \
					int 	card[3]; \
				} KHitTile_##type##_wrapper; \
			); \
		bool raw_added_ktile_KHitTile_##type = false;
#else
	#ifdef CTRL_FPGA_KERNEL_FILE
		#ifndef CTRL_HOST_COMPILE
			#ifndef CTRL_FPGA_BIN_NAME
				#define hit_ktileNewType( type ) \
					typedef struct __attribute__((packed)) __attribute__((aligned(CTRL_FPGA_STRUCT_ALIGNMENT))) { \
						__global type	* data; \
						int	origAcumCard[4]; \
						int 	card[3]; \
					} KHitTile_##type;
			#endif
		#endif
	#else
		#define hit_ktileNewType( type ) \
			typedef struct __attribute__((packed)) __attribute__((aligned(CTRL_FPGA_STRUCT_ALIGNMENT))) { \
				type	* data; \
				int	origAcumCard[4]; \
				int 	card[3]; \
			} KHitTile_##type;
	#endif
#endif

#define hit_ktileSelect1( ref, begin0, size0 ) \
	{ \
		ref.data+(begin0), \
		{ \
			ref.origAcumCard[0], 0, 0, 0 \
		}, \
		ref.card[0], 1, 1 \
	}


#define hit_ktileSelect2( ref, begin0, size0, begin1, size1 ) \
	{ \
		ref.data+(begin0 * ref.origAcumCard[0] + begin1), \
		{ \
			ref.origAcumCard[0], ref.origAcumCard[1], 0 \
		}, \
		ref.card[0], ref.card[1], 1 \
	}


#define hit_ktileSelect3( ref, begin0, size0, begin1, size1, begin2, size2 ) \
	{ \
		ref.data+(begin0 * ref.origAcumCard[0] * ref.origAcumCard[1] + begin1 * ref.origAcumCard[1] + begin2 ), \
		{ \
			ref.origAcumCard[0], ref.origAcumCard[1], ref.origAcumCard[2] \
		}, \
		ref.card[0], ref.card[1], ref.card[2] \
	}


#endif
