/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CPU_KERNELPROTO_H_
#define _CTRL_CPU_KERNELPROTO_H_

#include <stdio.h>

#include "Kernel/Ctrl_KernelArgs.h"
#include "Kernel/Ctrl_Thread.h"

//TODO @waxa lo de la funcion anterior que cargue el kernel y todo eso
#define CTRL_KERNEL_CPU( name, type, ... ) \
	void CTRL_KERNEL_##type##_##name ( Ctrl_Thread thread_id, __VA_ARGS__ )

// #define CTRL_KERNEL_Generic(name,type, ...)		\
// 	__kernel void CTRL_KERNEL_##type##_##name ( Ctrl_Thread threadId, __VA_ARGS__ )

// Punto de entrada al lanzamiento (equivalente a global en cuda) (hay que guardarlo en una variable o argo)
#define LAUNCH_CPU(name,  ...)				\
    void launch_Cpu_##name( Ctrl_Thread threads, CTRL_KERNEL_TYPED_KTILE( NULL, __VA_ARGS__ ) ) { }


#define CTRL_KERNEL_WRAP_CPU( name, argsList, type, ... ) \
    { \
        if (threads.z > 1) {  /* Executing in parallel the 3D threads */ \
            /*TODO @arturo numero de threads (depende del rango) */ \
            int i_aux; /*  private(i_aux, j_aux, k_aux)*/ \
            _Pragma("omp parallel for num_threads(request.cpu.n_cores)")  \
            for (i_aux = 0; i_aux < threads.x; i_aux++) { \
                for (int j_aux = 0 ; j_aux < threads.y; j_aux++) { \
                    for( int k_aux = 0; k_aux < threads.z; k_aux++) { \
                        Ctrl_Thread thread_id; \
                        thread_id.x = i_aux; \
                        thread_id.y = j_aux; \
                        thread_id.z = k_aux; \
                        CTRL_KERNEL_CPU_##name( thread_id, CTRL_KERNEL_ARG_LIST_ACCESS_KTILE( argsList, __VA_ARGS__ ) ); \
                    } \
                } \
            } \
        } else if (threads.y > 1) { /* Executing in parallel the 2D threads */ \
            int i_aux; /*  private(i_aux, j_aux, k_aux)*/ \
            _Pragma("omp parallel for num_threads(request.cpu.n_cores)")  \
            for (i_aux = 0; i_aux < threads.x; i_aux++) { \
                for (int j_aux = 0 ; j_aux < threads.y; j_aux++) { \
                    Ctrl_Thread thread_id; \
                    thread_id.x = i_aux; \
                    thread_id.y = j_aux; \
                    thread_id.z = 0; \
                    CTRL_KERNEL_CPU_##name( thread_id, CTRL_KERNEL_ARG_LIST_ACCESS_KTILE( argsList, __VA_ARGS__ ) ); \
                } \
            } \
        } else { /* Executing in parallel the 1D threads */ \
            int i_aux; /*  private(i_aux, j_aux, k_aux)*/ \
            _Pragma("omp parallel for num_threads(request.cpu.n_cores)") \
            for (i_aux = 0; i_aux < threads.x; i_aux++) { \
                Ctrl_Thread thread_id; \
                thread_id.x = i_aux; \
                CTRL_KERNEL_CPU_##name( thread_id, CTRL_KERNEL_ARG_LIST_ACCESS_KTILE( argsList, __VA_ARGS__ ) ); \
            } \
        } \
    }

#endif //_CTRL_CPU_KERNELPROTO_H_
