/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_KERNEL_CUDA_ARGS_H_
#define _CTRL_KERNEL_CUDA_ARGS_H_

#include <cuda_runtime.h>

#ifdef _CTRL_ARCH_CUDA_LEGACY_
	#include "Architectures/Cuda/Legacy/Ctrl_Cuda_Tile.h"
#else
	#include "Architectures/Cuda/Ctrl_Cuda_Tile.h"
#endif // _CTRL_ARCH_CUDA_LEGACY_
/*
 * Macros used to allow the use of KTiles and the GPU Kernel charazterization
 */
// TODO @waxa esto puede ir en el proto y KernelArgs seria completamente
// independiente

// FIXME: Needing refactor or cleanup -@ismael at 9/10/2018 14:48:39
// error: a value of type "void *" cannot be assigned to an entity of type
// "float *"
#define CTRL_KERNEL_CUDA_KTILE_DEVICE_DATA(name, type)                         \
    case CTRL_TYPE_CUDA:                                                       \
        ((KHitTile *)&k_##name)->data =                                        \
            (((Ctrl_Cuda_Tile *)name->ext)->p_device_data);                    \
        break;

#define CTRL_KERNEL_CUDA_STATIC 1
#define CTRL_KERNEL_CUDA_DYNAMIC 2

#define CTRL_KERNEL_CUDA_CHAR_STATIC(name, dimBlock, A, B, C)                  \
    int CTRL_KERNEL_CUDA_CHAR_type_##name = CTRL_KERNEL_CUDA_STATIC;           \
    dim3 CTRL_KERNEL_CUDA_CHAR_block_##name[CTRL_CUDA_NARCHS] =                \
        CTRL_KERNEL_CUDA_CHAR_##dimBlock##A##B##C;

#define CTRL_KERNEL_CUDA_CHAR_threads(name, arch)                              \
    (CTRL_KERNEL_CUDA_CHAR_type_##name == CTRL_KERNEL_CUDA_STATIC              \
         ? CTRL_KERNEL_CUDA_CHAR_block_##name[arch]                            \
         : CTRL_KERNEL_CUDA_CHAR_block_##name[0])

#define CTRL_KERNEL_CUDA_CHAR_grid(name, arch, threads)                        \
    dim3(((threads.dims > 1 ? threads.y:threads.x) + CTRL_KERNEL_CUDA_CHAR_threads(name, arch).x - 1) /       \
             CTRL_KERNEL_CUDA_CHAR_threads(name, arch).x,                      \
         ((threads.dims > 1 ? threads.x:threads.y) + CTRL_KERNEL_CUDA_CHAR_threads(name, arch).y - 1) /       \
             CTRL_KERNEL_CUDA_CHAR_threads(name, arch).y,                      \
         (threads.z + CTRL_KERNEL_CUDA_CHAR_threads(name, arch).z - 1) /       \
             CTRL_KERNEL_CUDA_CHAR_threads(name, arch).z);

#define CTRL_KERNEL_CUDA_KERNEL_CHAR(name, type, dims, ...)                    \
    CTRL_KERNEL_CUDA_KERNEL_CHAR_##type(name, dims, __VA_ARGS__);

/* Manual kernel characterization (type:1) */
#define CTRL_KERNEL_CUDA_KERNEL_CHAR_1(name, dims, ...)                        \
    int CTRL_KERNEL_CUDA_CHAR_type_##name = CTRL_KERNEL_CUDA_DYNAMIC;          \
    dim3 CTRL_KERNEL_CUDA_CHAR_block_##name[] = { CTRL_KERNEL_CUDA_KERNEL_CHAR_1_##dims(name, __VA_ARGS__) }; 

#define CTRL_KERNEL_CUDA_KERNEL_CHAR_1_1(name, X) CTRL_KERNEL_CUDA_KERNEL_CHAR_1_2(name, X, 1)

#define CTRL_KERNEL_CUDA_KERNEL_CHAR_1_2(name, X, Y) CTRL_KERNEL_CUDA_KERNEL_CHAR_1_3(name, X, Y, 1)

#define CTRL_KERNEL_CUDA_KERNEL_CHAR_1_3(name, X, Y, Z) dim3(X, Y, Z)

/* Automatic kernel characterization (type:2) */
#define CTRL_KERNEL_CUDA_KERNEL_CHAR_2(name, dims, ...)                        \
    CTRL_KERNEL_CUDA_CHAR_STATIC(name, dimBlock, __VA_ARGS__)

#endif // _CTRL_KERNEL_CUDA_ARGS_H_
