/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_KERNEL_CUDA_CHAR_H_
#define _CTRL_KERNEL_CUDA_CHAR_H_

#include <cuda_runtime.h>

#define CTRL_KERNEL_CUDA_NARCHS 4 // Número de arquitecturas que existen

/* Arquitecturas soportadas por la biblioteca */
#define CTRL_KERNEL_CUDA_ARCH_FERMI 2
#define CTRL_KERNEL_CUDA_ARCH_KEPLER 3

/*
 * Hilos por bloque para cada una de los tipos de kernels.
 * Cada uno del los valores corresponde con una arquitectura.
 */
#define CTRL_KERNEL_CUDA_CHAR_1defdefdef                                       \
    { dim3(0), dim3(0), dim3(256), dim3(256) }
#define CTRL_KERNEL_CUDA_CHAR_2defdefdef                                       \
    { dim3(0, 0), dim3(0, 0), dim3(256, 1), dim3(256, 1) }
#define CTRL_KERNEL_CUDA_CHAR_1mediumlowlow                                    \
    { dim3(128), dim3(128), dim3(128), dim3(128) }
#define CTRL_KERNEL_CUDA_CHAR_1fulllowlow                                      \
    { dim3(256), dim3(256), dim3(256), dim3(256) }
#define CTRL_KERNEL_CUDA_CHAR_2fulllowlow                                      \
    { dim3(128, 2), dim3(128, 2), dim3(128, 2), dim3(128, 2) }
#define CTRL_KERNEL_CUDA_CHAR_2square32lowlow                                  \
    { dim3(32, 32), dim3(32, 32), dim3(32, 32), dim3(32, 32) }
#define CTRL_KERNEL_CUDA_CHAR_2fullmediummedium                                \
    { dim3(128, 2), dim3(128, 2), dim3(128, 2), dim3(128, 2) }
#define CTRL_KERNEL_CUDA_CHAR_2fulllowhigh                                     \
    { dim3(0, 0), dim3(0, 0), dim3(64, 3), dim3(32, 4) }
#define CTRL_KERNEL_CUDA_CHAR_2mediummediummedium                              \
    { dim3(32, 4), dim3(32, 4), dim3(128, 2), dim3(128, 2) }
#define CTRL_KERNEL_CUDA_CHAR_2fixedsquare32                                   \
    { dim3(16, 16), dim3(16, 16), dim3(32, 32), dim3(32, 32) }
#define CTRL_KERNEL_CUDA_CHAR_2fixedsquare16                                   \
    { dim3(16, 16), dim3(16, 16), dim3(16, 16), dim3(16, 16) }
#define CTRL_KERNEL_CUDA_CHAR_2fixedsquare4                                    \
    { dim3(4, 4), dim3(4, 4), dim3(4, 4), dim3(4, 4) }
#define CTRL_KERNEL_CUDA_CHAR_2fixedsquare2                                    \
    { dim3(2, 2), dim3(2, 2), dim3(2, 2), dim3(2, 2) }

#endif // _CTRL_KERNEL_CUDA_CHAR_H_
