/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_CUDA_KERNELPROTO_H_
#define _CTRL_CUDA_KERNELPROTO_H_

#include "Kernel/Ctrl_KernelArgs.h"
#include "Kernel/Ctrl_Thread.h"

#include "Core/Ctrl_Request.h"

#include "Architectures/Cuda/Ctrl_Cuda_Helper.h"
#include "Architectures/Cuda/Ctrl_Cuda_Request.h"
#include "Architectures/Cuda/Ctrl_Cuda.h"

#define CTRL_KERNEL_CUDA(name, type, ...)                             \
    __device__ void Ctrl_Kernel_Cuda_##type##_##name(                 \
        Ctrl_Thread thread_id, CTRL_KERNEL_EXTRACT_ARGS( __VA_ARGS__ ) ) \
    {                                                                 \
        CTRL_KERNEL_EXTRACT_KERNEL_NO_STR(__VA_ARGS__)                \
    }

#define CTRL_KERNEL_CUDA_GENERIC(name, type, ...) \
    CTRL_KERNEL_CUDA(name, type, __VA_ARGS__)

#define CTRL_KERNEL_CUDA_LIB(name, type, ...)                                    \
    void Ctrl_Kernel_Cuda_##type##_##name(CTRL_KERNEL_EXTRACT_ARGS(__VA_ARGS__)) \
    {                                                                            \
        CTRL_KERNEL_EXTRACT_KERNEL_NO_STR(__VA_ARGS__)                           \
    }

#define LAUNCH_CUDA(name, ...)                                              \
    __global__ void launch_Cuda_##name(                                     \
        Ctrl_Thread threads, CTRL_KERNEL_TYPED_KTILE(NULL, __VA_ARGS__))    \
    {                                                                       \
        int thread_idx = blockIdx.x * blockDim.x + threadIdx.x;                \
        int thread_idy = blockIdx.y * blockDim.y + threadIdx.y;                \
        int thread_idz = blockIdx.z * blockDim.z + threadIdx.z;                \
        if (threads.dims > 1)                                                  \
        {                                                                      \
            int thread_temp = thread_idx;                                      \
            thread_idx = thread_idy;                                           \
            thread_idy = thread_temp;                                          \
        }                                                                      \
        if (thread_idx >= threads.x || thread_idy >= threads.y ||              \
            thread_idz >= threads.z)                                           \
            return;                                                            \
        Ctrl_Thread thread_id = {threads.dims, thread_idx, thread_idy,         \
                                 thread_idz};                               \
        Ctrl_Kernel_Cuda_CUDA_##name(thread_id,                             \
                                     CTRL_KERNEL_VALUE(NULL, __VA_ARGS__)); \
    }

#define LAUNCH_CUDA_GENERIC(name, ...)                                         \
    __global__ void launch_Cuda_Generic_##name(                                \
        Ctrl_Thread threads, CTRL_KERNEL_TYPED_KTILE(NULL, __VA_ARGS__))       \
    {                                                                          \
        int thread_idx = blockIdx.x * blockDim.x + threadIdx.x;                \
        int thread_idy = blockIdx.y * blockDim.y + threadIdx.y;                \
        int thread_idz = blockIdx.z * blockDim.z + threadIdx.z;                \
        if (threads.dims > 1)                                                  \
        {                                                                      \
            int thread_temp = thread_idx;                                      \
            thread_idx = thread_idy;                                           \
            thread_idy = thread_temp;                                          \
        }                                                                      \
        if (thread_idx >= threads.x || thread_idy >= threads.y ||              \
            thread_idz >= threads.z)                                           \
            return;                                                            \
        Ctrl_Thread thread_id = {threads.dims, thread_idx, thread_idy,         \
                                 thread_idz};                                  \
        Ctrl_Kernel_Cuda_GENERIC_##name(thread_id,                             \
                                        CTRL_KERNEL_VALUE(NULL, __VA_ARGS__)); \
    }

#define LAUNCH_CUDA_LIB(...)

#define CTRL_KERNEL_WRAP_CUDA(name, argsList, type, ...)                       \
    {                                                                          \
        dim3 grid = CTRL_KERNEL_CUDA_CHAR_grid(                                \
            name, CTRL_KERNEL_CUDA_ARCH_KEPLER, threads);                      \
        dim3 block =                                                           \
            CTRL_KERNEL_CUDA_CHAR_threads(name, CTRL_KERNEL_CUDA_ARCH_KEPLER); \
        launch_Cuda_##name<<<grid, block, 0, *(request.cuda.stream)>>>(        \
            threads,                                                           \
            CTRL_KERNEL_ARG_LIST_ACCESS_KTILE(argsList, __VA_ARGS__));         \
        CUDA_ERROR(); \
    };

#define CTRL_KERNEL_WRAP_CUDA_GENERIC(name, argsList, type, ...)               \
    {                                                                          \
        dim3 grid = CTRL_KERNEL_CUDA_CHAR_grid(                                \
            name, CTRL_KERNEL_CUDA_ARCH_KEPLER, threads);                      \
        dim3 block =                                                           \
            CTRL_KERNEL_CUDA_CHAR_threads(name, CTRL_KERNEL_CUDA_ARCH_KEPLER);  \
        launch_Cuda_Generic_##name<<<grid, block, 0, *(request.cuda.stream)>>>( \
            threads,                                                           \
            CTRL_KERNEL_ARG_LIST_ACCESS_KTILE(argsList, __VA_ARGS__));         \
        CUDA_ERROR(); \
    };

#define CTRL_KERNEL_WRAP_CUDA_LIB(name, argsList, type, ...)           \
    {                                                                  \
        Ctrl_Kernel_Cuda_##type##_##name(                              \
            CTRL_KERNEL_ARG_LIST_ACCESS_KTILE(argsList, __VA_ARGS__)); \
    };

#endif // _CTRL_CUDA_KERNELPROTO_H_
