/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#ifndef _CTRL_OPENCL_KERNELPROTO_H_
#define _CTRL_OPENCL_KERNELPROTO_H_

#include <string.h>

#ifndef CL_USE_DEPRECATED_OPENCL_1_2_APIS
#define CL_USE_DEPRECATED_OPENCL_1_2_APIS
#endif // CL_USE_DEPRECATED_OPENCL_1_2_APIS

#include <CL/cl.h>

#include "Kernel/Ctrl_KernelArgs.h"
#include "Kernel/Ctrl_Thread.h"

#include "Core/Ctrl_Request.h"

#include "Architectures/OpenCL/Ctrl_OpenCL_Request.h"
#include "Architectures/OpenCL/Ctrl_OpenCL_Gpu.h"


#define CTRL_KERNEL_OPENCL_PARSE_ARGS_INVAL( type, name )   " const " , CTRL_KERNEL_STRINGIFY( type ) , " " , CTRL_KERNEL_STRINGIFY( name ) 
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_IN( type, name )      " const " , CTRL_KERNEL_STRINGIFY( K##type##_wrapper ) , " ", CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name ) , " , __global const " , raw_ktile_K##type , " " ,  CTRL_KERNEL_STRINGIFY( *ctrl_mem_wrapper_##name ) 
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_OUT( type, name )     "  " , CTRL_KERNEL_STRINGIFY( K##type##_wrapper ) , " ",  CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name ) , " , __global " , raw_ktile_K##type  , " " , CTRL_KERNEL_STRINGIFY( *ctrl_mem_wrapper_##name ) 
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_IO( type, name ) CTRL_KERNEL_OPENCL_PARSE_ARGS_OUT( type, name )

#define CTRL_KERNEL_OPENCL_PARSE_ARGS_1( role, type, name ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name)
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_2( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_1( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_3( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_2( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_4( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_3( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_5( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_4( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_6( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_5( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_7( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_6( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_8( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_7( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_9( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_8( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_10( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_9( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_11( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_10( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_12( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_11( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_13( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_12( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_14( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_13( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_15( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_14( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_16( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_15( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_17( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_16( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_18( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_17( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_19( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_18( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_20( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_19( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_ARGS_21( role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##role( type, name ) , CTRL_KERNEL_OPENCL_PARSE_ARGS_20( __VA_ARGS__ )

#define CTRL_KERNEL_OPENCL_PARSE_ARGS( n, ... ) CTRL_KERNEL_OPENCL_PARSE_ARGS_##n( __VA_ARGS__ ) 


#define CTRL_KERNEL_OPENCL_RESET_TILES_INVAL( type, name )   
#define CTRL_KERNEL_OPENCL_RESET_TILES_IN( type, name )     raw_added_ktile_K##type = false;
#define CTRL_KERNEL_OPENCL_RESET_TILES_OUT( type, name )    CTRL_KERNEL_OPENCL_RESET_TILES_IN( type, name ) 
#define CTRL_KERNEL_OPENCL_RESET_TILES_IO( type, name )     CTRL_KERNEL_OPENCL_RESET_TILES_OUT( type, name )

#define CTRL_KERNEL_OPENCL_RESET_TILES_1( role, type, name ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name)
#define CTRL_KERNEL_OPENCL_RESET_TILES_2( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_1( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_3( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_2( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_4( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_3( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_5( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_4( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_6( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_5( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_7( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_6( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_8( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_7( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_9( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_8( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_10( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_9( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_11( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_10( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_12( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_11( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_13( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_12( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_14( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_13( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_15( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_14( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_16( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_15( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_17( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_16( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_18( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_17( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_19( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_18( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_20( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_19( __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_RESET_TILES_21( role, type, name, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##role( type, name )  CTRL_KERNEL_OPENCL_RESET_TILES_20( __VA_ARGS__ )

#define CTRL_KERNEL_OPENCL_RESET_TILES( n, ... ) CTRL_KERNEL_OPENCL_RESET_TILES_##n( __VA_ARGS__ ) 


#define CTRL_KERNEL_OPENCL_PARSE_TILES_INVAL( string, type, name ) 
#define CTRL_KERNEL_OPENCL_PARSE_TILES_IN( string, type, name )     if (!raw_added_ktile_K##type) { strcat(string, " typedef unsigned char BYTE; "); strcat(string, raw_def_ktile_K##type ); raw_added_ktile_K##type = true; }
#define CTRL_KERNEL_OPENCL_PARSE_TILES_OUT( string, type, name ) CTRL_KERNEL_OPENCL_PARSE_TILES_IN( string, type, name )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_IO( string, type, name ) CTRL_KERNEL_OPENCL_PARSE_TILES_OUT( string, type, name )

#define CTRL_KERNEL_OPENCL_PARSE_TILES_1( string, role, type, name ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name)
#define CTRL_KERNEL_OPENCL_PARSE_TILES_2( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_1( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_3( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_2( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_4( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_3( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_5( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_4( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_6( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_5( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_7( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_6( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_8( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_7( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_9( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_8( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_10( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_9( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_11( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_10( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_12( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_11( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_13( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_12( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_14( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_13( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_15( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_14( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_16( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_15( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_17( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_16( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_18( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_17( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_19( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_18( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_20( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_19( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_PARSE_TILES_21( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_PARSE_TILES_20( string, __VA_ARGS__ )

#define CTRL_KERNEL_OPENCL_PARSE_TILES( string, n, ... ) CTRL_KERNEL_OPENCL_PARSE_TILES_##n( string, __VA_ARGS__ ) 

#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_INVAL( string, type, name ) 
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_IN( string, type, name ) \
		strcat(string, CTRL_KERNEL_STRINGIFY( K##type##_read ) ); strcat(string, " "); strcat(string, CTRL_KERNEL_STRINGIFY( name )); \
		strcat(string, " = { .data = "); strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_mem_wrapper_##name )); strcat(string, ", .origAcumCard = { "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[0], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[1], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[2], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[3] } , .card = { "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[0], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[1], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[2] }}; ");

#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_OUT( string, type, name ) \
		strcat(string, CTRL_KERNEL_STRINGIFY( K##type##_write ) ); strcat(string, " "); strcat(string, CTRL_KERNEL_STRINGIFY( name )); \
		strcat(string, " = { .data = "); strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_mem_wrapper_##name )); strcat(string, ", .origAcumCard = { "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[0], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[1], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[2], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".origAcumCard[3] } , .card = { "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[0], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[1], "); \
		strcat(string, CTRL_KERNEL_STRINGIFY( ctrl_ktile_wrapper_##name )); strcat(string, ".card[2] }}; ");

#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_IO( string, type, name ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_OUT( string, type, name )

#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_1( string, role, type, name ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name)
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_2( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_1( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_3( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_2( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_4( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_3( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_5( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_4( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_6( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_5( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_7( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_6( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_8( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_7( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_9( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_8( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_10( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_9( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_11( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_10( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_12( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_11( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_13( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_12( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_14( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_13( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_15( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_14( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_16( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_15( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_17( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_16( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_18( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_17( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_19( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_18( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_20( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_19( string, __VA_ARGS__ )
#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_21( string, role, type, name, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##role( string, type, name )  CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_20( string, __VA_ARGS__ )

#define CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES( string, n, ... ) CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES_##n( string, __VA_ARGS__ ) 


#define CTRL_KERNEL_OPENCL_PARSE_THREADS " Ctrl_Thread thread_id = { .dims = ctrl_thread.dims, .x = get_global_id(0), .y = get_global_id(1), .z = get_global_id(2) }; \
		if ( ctrl_thread.dims > 1 ) { \
			thread_id.x = get_global_id(1); \
			thread_id.y = get_global_id(0); \
		} \
		if ( thread_id.x >= ctrl_thread.x || thread_id.y >= ctrl_thread.y || thread_id.z >= ctrl_thread.z ) { \
			return ; \
		} " \


#define CTRL_KERNEL_OPENCL_GPU( name, type, ... ) \
		static cl_kernel *p_kernel_##type##_##name = NULL; \
		static cl_program *p_program_##type##_##name = NULL; \
		static char *p_kernel_raw_##type##_##name = NULL; \
		static const char *p_ctrl_kernel_name_##type##_##name = CTRL_KERNEL_STRINGIFY( CTRL_KERNEL_##type##_##name ) ; \
		static const char *p_ctrl_kernel_string_##type##_##name = CTRL_KERNEL_EXTRACT_KERNEL( __VA_ARGS__ ) ;

#define CTRL_KERNEL_OPENCL_GPU_GENERIC( name, type, ... ) \
		CTRL_KERNEL_OPENCL_GPU( name, type, __VA_ARGS__ )


#define CTRL_KERNEL_OPENCL_GPU_LIB( ... )


#define LAUNCH_OPENCL_GPU( name,  ... )

#define LAUNCH_OPENCL_GPU_GENERIC( name,  ... ) \
		LAUNCH_OPENCL_GPU( name,  __VA_ARGS__ )

#define LAUNCH_OPENCL_GPU_LIB( ... )

#define CTRL_KERNEL_WRAP_OPENCL_GPU( name, args_list, type, n_args, ... ) \
		{ \
			cl_int err; \
			\
			if (p_kernel_##type##_##name == NULL) { \
				p_kernel_##type##_##name = (cl_kernel *)malloc(sizeof(cl_kernel)); \
				p_program_##type##_##name = (cl_program *)malloc(sizeof(cl_program)); \
				\
				char *pp_args_names_##type##_##name[] = { " __kernel void " , CTRL_KERNEL_STRINGIFY( CTRL_KERNEL_##type##_##name ), " ( ", CTRL_KERNEL_OPENCL_PARSE_ARGS( n_args, __VA_ARGS__ ) } ; \
				p_kernel_raw_##type##_##name = (char *)malloc(10000 * sizeof(char)); \
				p_kernel_raw_##type##_##name[0] = '\0'; \
				CTRL_KERNEL_OPENCL_RESET_TILES( n_args, __VA_ARGS__ ); \
				CTRL_KERNEL_OPENCL_PARSE_TILES( p_kernel_raw_##type##_##name, n_args, __VA_ARGS__ ) ; \
				CTRL_KERNEL_OPENCL_RESET_TILES( n_args, __VA_ARGS__ ); \
				\
				strcat(p_kernel_raw_##type##_##name, CTRL_THREAD_STRINGIFY ); \
				\
				strcat(&p_kernel_raw_##type##_##name[0], &pp_args_names_##type##_##name[0][0]); \
				strcat(&p_kernel_raw_##type##_##name[0], &pp_args_names_##type##_##name[1][0]); \
				strcat(&p_kernel_raw_##type##_##name[0], &pp_args_names_##type##_##name[2][0]); \
				int index = 3; \
				for (int i = 0; i < request.opencl.n_arguments; i++) { \
					if (i != 0) { \
						strcat(p_kernel_raw_##type##_##name, " , "); \
					} \
					if (request.opencl.p_roles[i] == KERNEL_INVAL) { \
						for (int j = 0; j < 4; j++) { \
							strcat(p_kernel_raw_##type##_##name, pp_args_names_##type##_##name[index++]); \
						} \
					} else { \
						for (int j = 0; j < 8; j++) { \
							strcat(p_kernel_raw_##type##_##name, pp_args_names_##type##_##name[index++]); \
						} \
					} \
				}  \
				int init_kernel = 0; \
				while (true) { \
					if (p_ctrl_kernel_string_##type##_##name[init_kernel] == '{' ) { \
						init_kernel++; \
						break; \
					} \
					init_kernel++; \
				} \
				strcat(p_kernel_raw_##type##_##name, " , const Ctrl_Thread ctrl_thread ) { "); \
				CTRL_KERNEL_OPENCL_MOUNT_INNER_TILES( p_kernel_raw_##type##_##name, n_args, __VA_ARGS__ ); \
				strcat(p_kernel_raw_##type##_##name, CTRL_KERNEL_OPENCL_PARSE_THREADS); \
				strcat(p_kernel_raw_##type##_##name, &p_ctrl_kernel_string_##type##_##name[init_kernel] ); \
				const size_t kernel_raw_size_##type##_##name = strlen(p_kernel_raw_##type##_##name); \
				\
				*p_program_##type##_##name = clCreateProgramWithSource(*(request.opencl.context), 1, \
						(const char **)(&p_kernel_raw_##type##_##name), &kernel_raw_size_##type##_##name, &err); \
				OPENCL_ASSERT_ERROR( err ); \
				err = clBuildProgram(*p_program_##type##_##name, 1, request.opencl.device_id, NULL, NULL, NULL); \
				if (err == CL_BUILD_PROGRAM_FAILURE) {\
					printf("\n\n%s\n\n", p_kernel_raw_##type##_##name); fflush(stdout); \
					size_t log_size; \
					clGetProgramBuildInfo(*p_program_##type##_##name, *(request.opencl.device_id), CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size); \
					char *log = (char *) malloc(log_size); \
					clGetProgramBuildInfo(*p_program_##type##_##name, *(request.opencl.device_id), CL_PROGRAM_BUILD_LOG, log_size, log, NULL); \
					printf("%s\n", log); fflush(stdout); \
					free(log); \
				}  \
				OPENCL_ASSERT_ERROR( err ); \
				/* OPENCL_ASSERT_OP( clBuildProgram(*p_program_##type##_##name, 1, request.opencl.device_id, NULL, NULL, NULL) ); */ \
				\
				*p_kernel_##type##_##name = clCreateKernel(*p_program_##type##_##name, (const char *)p_ctrl_kernel_name_##type##_##name, &err); \
				OPENCL_ASSERT_ERROR( err );  \
				Ctrl_OpenCL_Res *p_res = (Ctrl_OpenCL_Res *)malloc(sizeof(Ctrl_OpenCL_Res)); \
				p_res->p_kernel = p_kernel_##type##_##name; \
				p_res->p_program = p_program_##type##_##name; \
				p_res->p_kernel_raw = p_kernel_raw_##type##_##name; \
				p_res->p_next = NULL; \
				if (request.opencl.p_res_list_head == NULL) { \
					request.opencl.p_res_list_head = p_res; \
				} else { \
					Ctrl_OpenCL_Res *p_aux = request.opencl.p_res_list_head; \
					while (p_aux->p_next != NULL) { \
						p_aux = p_aux->p_next; \
					} \
					p_aux->p_next = p_res; \
				} \
			} \
			err = 0; \
			int arg_pos = 0; \
			for (int i = 0; i < request.opencl.n_arguments; i++) { \
				if (request.opencl.p_roles[i] == KERNEL_INVAL) { \
					err |= clSetKernelArg(*p_kernel_##type##_##name, arg_pos, \
							request.opencl.p_displacements[i+1] - request.opencl.p_displacements[i], \
							((uint8_t *)args_list + request.opencl.p_displacements[i])); \
					arg_pos++; \
				} else { \
					KHitTile *p_ktile = (KHitTile *)((uint8_t *)args_list + request.opencl.p_displacements[i]); \
					KHitTile_opencl_wrapper ktile_opencl_wraper = { .origAcumCard = { p_ktile->origAcumCard[0], \
							p_ktile->origAcumCard[1], p_ktile->origAcumCard[2], p_ktile->origAcumCard[3] }, \
							.card = { p_ktile->card[0], p_ktile->card[1], p_ktile->card[2] }}; \
					err |= clSetKernelArg(*p_kernel_##type##_##name, arg_pos, \
							sizeof(KHitTile_opencl_wrapper), &ktile_opencl_wraper); \
					arg_pos++; \
					err |= clSetKernelArg(*p_kernel_##type##_##name, arg_pos, \
							sizeof(cl_mem), (cl_mem *)(p_ktile->data)); \
					arg_pos++; \
				} \
			} \
			err |= clSetKernelArg(*p_kernel_##type##_##name, arg_pos, \
					sizeof(Ctrl_Thread), &threads); \
			arg_pos++; \
			size_t global_size[3]; \
			if (threads.dims == 1) { \
				global_size[0] = threads.x; \
				global_size[1] = threads.y; \
				global_size[2] = threads.z; \
			} else { \
				global_size[0] = threads.y; \
				global_size[1] = threads.x; \
				global_size[2] = threads.z; \
			} \
			if (threads.dims >= 1) { \
				if ((global_size[0] % local_size_OPENCL_GPU_##name[0]) != 0) { \
					global_size[0] += (local_size_OPENCL_GPU_##name[0] - (global_size[0] % local_size_OPENCL_GPU_##name[0])); \
				} \
			} \
            if (threads.dims >= 2) { \
				if ((global_size[1] % local_size_OPENCL_GPU_##name[1]) != 0) { \
					global_size[1] += (local_size_OPENCL_GPU_##name[1] - (global_size[1] % local_size_OPENCL_GPU_##name[1])); \
				} \
			} \
            if (threads.dims == 3) { \
				if ((global_size[2] % local_size_OPENCL_GPU_##name[2]) != 0) { \
					global_size[2] += (local_size_OPENCL_GPU_##name[2] - (global_size[2] % local_size_OPENCL_GPU_##name[2])); \
				} \
			} \
			/* printf("threads -> dims: %d, x: %d, y: %d, z: %d\n", threads.dims, threads.x, threads.y, threads.z); fflush(stdout); */ \
        	/* printf("local sizes: %lu, %lu, %lu\n", local_size_OPENCL_GPU_##name[0], local_size_OPENCL_GPU_##name[1], local_size_OPENCL_GPU_##name[2]); */ \
        	/* printf("global sizes: %lu, %lu, %lu\n", global_size[0], global_size[1], global_size[2]); */ \
			OPENCL_ASSERT_OP( clEnqueueNDRangeKernel(*(request.opencl.queue), *p_kernel_##type##_##name , \
					threads.dims, 0, global_size, local_size_OPENCL_GPU_##name, \
					request.opencl.n_event_wait, request.opencl.p_event_wait_list, request.opencl.p_last_kernel_event) ); \
			OPENCL_ASSERT_OP( clFlush(*(request.opencl.queue)) ); \
		} 

#define CTRL_KERNEL_WRAP_OPENCL_GPU_GENERIC( name, args_list, type, n_args, ... ) \
		CTRL_KERNEL_WRAP_OPENCL_GPU( name, args_list, type, n_args, __VA_ARGS__ ) 

	
#define CTRL_KERNEL_WRAP_OPENCL_GPU_LIB( ... )


	// if (err == CL_BUILD_PROGRAM_FAILURE) { \
	// 			size_t log_size; \
	// 			clGetProgramBuildInfo(*p_program_##type##_##name, request.opencl.device_id, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size); \
	// 			char *log = (char *) malloc(log_size); \
	// 			clGetProgramBuildInfo(*p_program_##type##_##name, request.opencl.device_id, CL_PROGRAM_BUILD_LOG, log_size, log, NULL); \
	// 			printf("%s\n", log); \
	// 			free(log); \
	// 		} \

#endif //_CTRL_OPENCL_KERNELPROTO_H_
