/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
/*
 * CAL_KProto: Macros to generate the code, from the prototype declaration of a kernel,
 * 			of the enqueue/dequeue(launching) functions. The launching function is a
 * 			wrapper to invocate the actual kernel implementations
 *
 * @version 1.2
 * @author Ana Moreton-Fernandez
 * @date OCt 2017
 *
 * NOTE: Macros to support up to 7 different implementations, and up to 9 different parameters
 *		in a kernel call. Add more macros for higher number of implementations and/or parameters.
 */

#ifndef _CTRL_KERNEL_PROTO_H_
#define _CTRL_KERNEL_PROTO_H_

#ifdef CTRL_HOST_COMPILE
	#define MODEL_TASK	0
	#define MODEL_NDRANGE	1
#endif

#ifndef CTRL_FPGA_KERNEL_FILE
/* struct kernel_params {
	cl_kernel *kernel;
	cl_program *program;
	const char *kernel_name;
	unsigned char *binary_str;
	size_t binary_length;

	struct kernel_params *next;
} kernel_params; */

#include "Kernel/Ctrl_ImplType.h"
#include "Kernel/Ctrl_KernelArgs.h"

#include "Core/Ctrl_TaskQueue.h"
#endif

#ifdef _CTRL_ARCH_CPU_
	#include "Kernel/Architectures/Cpu/Ctrl_Cpu_KernelProto.h"
#else
	#define CTRL_KERNEL_CPU( ... )
	#define CTRL_KERNEL_WRAP_CPU( ... )
	#define LAUNCH_CPU( ... )
	#define CTRL_KERNEL_CPU_LIB( ... )
	#define CTRL_KERNEL_WRAP_CPU_LIB( ... )
	#define LAUNCH_CPU_LIB( ... )
	#define CTRL_KERNEL_CPU_GENERIC( ... )
	#define CTRL_KERNEL_WRAP_CPU_GENERIC( ... )
	#define LAUNCH_CPU_GENERIC( ... )
#endif // _CTRL_ARCH_CPU_

#ifdef _CTRL_ARCH_CUDA_
	#include "Kernel/Architectures/Cuda/Ctrl_Cuda_KernelProto.h"
#else
	#define CTRL_KERNEL_CUDA( ... )
	#define CTRL_KERNEL_WRAP_CUDA( ... )
	#define LAUNCH_CUDA( ... )
	#define CTRL_KERNEL_CUDA_LIB( ... )
	#define CTRL_KERNEL_WRAP_CUDA_LIB( ... )
	#define LAUNCH_CUDA_LIB( ... )
	#define CTRL_KERNEL_CUDA_GENERIC( ... )
	#define CTRL_KERNEL_WRAP_CUDA_GENERIC( ... )
	#define LAUNCH_CUDA_GENERIC( ... )
#endif

#ifdef _CTRL_ARCH_OPENCL_GPU_
	#include "Kernel/Architectures/OpenCL/Ctrl_OpenCL_KernelProto.h"
#else
	#define CTRL_KERNEL_OPENCL_GPU( ... )
	#define CTRL_KERNEL_WRAP_OPENCL_GPU( ... )
	#define LAUNCH_OPENCL_GPU( ... )
	#define CTRL_KERNEL_OPENCL_GPU_LIB( ... )
	#define CTRL_KERNEL_WRAP_OPENCL_GPU_LIB( ... )
	#define LAUNCH_OPENCL_GPU_LIB( ... )
	#define CTRL_KERNEL_OPENCL_GPU_GENERIC( ... )
	#define CTRL_KERNEL_WRAP_OPENCL_GPU_GENERIC( ... )
	#define LAUNCH_OPENCL_GPU_GENERIC( ... )
#endif // _CTRL_ARCH_OPENCL_GPU_

#ifdef _CTRL_ARCH_FPGA_
	#include "Kernel/Architectures/FPGA/Ctrl_FPGA_KernelProto.h"
#else
	#define CTRL_KERNEL_FPGA( ... )
	#define CTRL_KERNEL_WRAP_FPGA( ... )
	#define LAUNCH_FPGA( ... )
	#define CTRL_KERNEL_FPGA_LIB( ... )
	#define CTRL_KERNEL_WRAP_FPGA_LIB( ... )
	#define LAUNCH_FPGA_LIB( ... )
	#define CTRL_KERNEL_FPGA_GENERIC( ... )
	#define CTRL_KERNEL_WRAP_FPGA_GENERIC( ... )
	#define LAUNCH_FPGA_GENERIC( ... )
#endif // _CTRL_ARCH_FPGA_

#ifdef _CTRL_ARCH_XPHI_
	#include "Kernel/Architectures/XPhi/Ctrl_XPhi_KernelProto.h"
#else
	#define CTRL_KERNEL_XPHI( ... )
	#define CTRL_KERNEL_WRAP_XPHI( ... )
	#define LAUNCH_XPHI( ... )
	#define CTRL_KERNEL_XPHI_LIB( ... )
	#define CTRL_KERNEL_WRAP_XPHI_LIB( ... )
	#define LAUNCH_XPHI_LIB( ... )
	#define CTRL_KERNEL_XPHI_GENERIC( ... )
	#define CTRL_KERNEL_WRAP_XPHI_GENERIC( ... )
	#define LAUNCH_XPHI_GENERIC( ... )
#endif // _CTRL_ARCH_XPHI_

/*
 *******************************************************************************************
 ****************************************  KERNEL  *****************************************
 *******************************************************************************************
 */

#define CTRL_KERNEL_GENERIC( ... ) \
		CTRL_KERNEL_CPU_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_CUDA_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_OPENCL_GPU_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_FPGA_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_XPHI_GENERIC( __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_GENERIC( ... ) \
		CTRL_KERNEL_WRAP_CPU_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_WRAP_CUDA_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_WRAP_OPENCL_GPU_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_WRAP_FPGA_GENERIC( __VA_ARGS__ ) \
		CTRL_KERNEL_WRAP_XPHI_GENERIC( __VA_ARGS__ )

#define LAUNCH_GENERIC( ... ) \
		LAUNCH_CPU_GENERIC( __VA_ARGS__ ) \
		LAUNCH_CUDA_GENERIC( __VA_ARGS__ ) \
		LAUNCH_OPENCL_GPU_GENERIC( __VA_ARGS__ ) \
		LAUNCH_FPGA_GENERIC( __VA_ARGS__ ) \
		LAUNCH_XPHI_GENERIC( __VA_ARGS__ )


#define CTRL_KERNEL( name, type, ... ) \
	CTRL_KERNEL_##type( name, type, __VA_ARGS__ ) 


/* Case structure to call/launch one of the kernel implementations */
#define CTRL_KERNEL_WRAP_LAUNCH_1( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList, type, __VA_ARGS__ ) \
	break; 

#define CTRL_KERNEL_WRAP_LAUNCH_2( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList, type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 1, __VA_ARGS__ ) ) \
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_1( name, argsList, __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_LAUNCH_3( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList, type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 2, __VA_ARGS__ ) ) \
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_2( name, argsList, __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_LAUNCH_4( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList, type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 3, __VA_ARGS__ ) ) \
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_3( name, argsList, __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_LAUNCH_5( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList, type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 4, __VA_ARGS__ ) ) ;\
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_4( name, argsList, __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_LAUNCH_6( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList,type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 5, __VA_ARGS__) ) ;\
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_5( name, argsList, __VA_ARGS__ )

#define CTRL_KERNEL_WRAP_LAUNCH_7( name, argsList, type, ... )	\
	case CTRL_IMPLTYPE_##type: CTRL_KERNEL_WRAP_##type( name, argsList,type, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 6, __VA_ARGS__) ) ;\
	break; \
	CTRL_KERNEL_WRAP_LAUNCH_6( name, argsList, __VA_ARGS__ )

#ifdef CTRL_FPGA_KERNEL_FILE
/* Case structure to define one of the kernel implementations */
#define CTRL_KERNEL_WRAP_DEFINE( name, type, ... )	\
	CTRL_KERNEL_WRAP_##type( name, type, __VA_ARGS__ )
#endif

/*
 * Macros to define the functions to launch kernels
 */
#define CTRL_KERNEL_LAUNCHER_1( name, type, ...)	\
		LAUNCH_##type (name, __VA_ARGS__)

#define CTRL_KERNEL_LAUNCHER_2(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 1, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_1(name, __VA_ARGS__) 

#define CTRL_KERNEL_LAUNCHER_3(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 2, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_2(name, __VA_ARGS__) 

#define CTRL_KERNEL_LAUNCHER_4(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 3, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_3(name, __VA_ARGS__) 

#define CTRL_KERNEL_LAUNCHER_5(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 4, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_4(name, __VA_ARGS__) 

#define CTRL_KERNEL_LAUNCHER_6(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 5, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_5(name, __VA_ARGS__) 

#define CTRL_KERNEL_LAUNCHER_7(name, type, ...)	\
		LAUNCH_##type (name, CTRL_KERNEL_SKIP_IMPL( ROLED, NULL, 6, __VA_ARGS__))	\
		CTRL_KERNEL_LAUNCHER_6(name, __VA_ARGS__) 


/*
 *******************************************************************************************
 ***************************************  MASKs   ******************************************
 *******************************************************************************************
 */

#define CTRL_KERNEL_MASK_1(type, ...)	\
	CTRL_IMPLTYPE_##type

#define CTRL_KERNEL_MASK_2(type, ...)	\
	CTRL_IMPLTYPE_##type + \
	CTRL_KERNEL_MASK_1( __VA_ARGS__)

#define CTRL_KERNEL_MASK_3(type, ...)	\
	type + CTRL_KERNEL_MASK_2( __VA_ARGS__)

#define CTRL_KERNEL_MASK_4(type, ...)	\
	type + CTRL_KERNEL_MASK_3( __VA_ARGS__)

#define CTRL_KERNEL_MASK_5(type, ...)	\
	type + CTRL_KERNEL_MASK_4( __VA_ARGS__)

#define CTRL_KERNEL_MASK_6(type, ...)	\
	type + CTRL_KERNEL_MASK_5( __VA_ARGS__)

#define CTRL_KERNEL_MASK_7(type, ...)	\
	type + CTRL_KERNEL_MASK_6( __VA_ARGS__)


/*
 *******************************************************************************************
 ***************************************  PROTO  *******************************************
 *******************************************************************************************
 */

/* 
 * Prototype and implementations declaration. 
 * Build the kernel task creation, and the launch wrapper function
 */
#ifndef CTRL_FPGA_KERNEL_FILE
	#define CTRL_KERNEL_PROTO( name, n_implementations, ... ) \
		\
		/* @author: Gabriel Rodriguez-Canal */ \
		/* @brief: kernel pointers belonging to CTRL_KERNEL in OpenCL backend must be placed here in FPGA version to use CTRL_KERNEL in .cl files */ \
		/* CTRL_KERNEL_LAUNCH_POINTERS( name, __VA_ARGS__ ) */ \
		\
		CTRL_INIT_KERNEL( name, __VA_ARGS__ ) \
		\
		CTRL_KERNEL_LAUNCHER_##n_implementations( name, __VA_ARGS__ ) \
		\
		void Ctrl_KernelWrapper_##name( Ctrl_Request request, Ctrl_ImplType impl_type, int device_id, Ctrl_Thread threads, void * args_list ) { \
			switch ( device_id ) { \
				CTRL_KERNEL_WRAP_LAUNCH_##n_implementations( name, args_list, __VA_ARGS__ ) \
				default: \
					fprintf(stderr, "CTrl Internal error: Wrong implementation type on launching wrapper: %s, %s[%d]\n", #name, __FILE__, __LINE__ ); \
			} \
		} \
		\
		Ctrl_Task Ctrl_KernelTaskCreate_##name( Ctrl_Type ctrl_type, Ctrl_Thread threads, CTRL_KERNEL_SKIP_IMPL( TYPED_POINTER, NULL, n_implementations, __VA_ARGS__ ) ) { \
			Ctrl_Task task = CTRL_TASK_NULL; \
			task.threads = threads; \
			task.task_type = CTRL_TASK_TYPE_KERNEL; \
			task.n_arguments = CTRL_KERNEL_SKIP_IMPL( ARG_COUNT_LIST_ELEMENTS, NULL, n_implementations, __VA_ARGS__ ); \
			task.pfn_kernel_wrapper = Ctrl_KernelWrapper_##name; \
			CTRL_KERNEL_SKIP_IMPL( ARG_LIST_CREATE_KTILE, task.p_arguments, n_implementations, __VA_ARGS__); \
			CTRL_KERNEL_SKIP_IMPL( ROLES, task.p_roles, n_implementations, __VA_ARGS__); \
			CTRL_KERNEL_SKIP_IMPL( POINTERS, task.pp_pointers, n_implementations, __VA_ARGS__); \
			CTRL_KERNEL_SKIP_IMPL( DISPLACEMENTS, task.p_displacements, n_implementations, __VA_ARGS__); \
			int sum_impl= CTRL_KERNEL_MASK_##n_implementations(__VA_ARGS__); \
			int dev = Ctrl_Dev(ctrl_type, sum_impl); \
			task.device_id = dev; \
			return task; \
		}
#else
	#define CTRL_KERNEL_PROTO( name, n_implementations, ... ) \
		CTRL_KERNEL_WRAP_DEFINE( name, __VA_ARGS__ )
#endif

#define CTRL_KERNEL_LAUNCH_POINTERS( name, type, ... ) \
	CTRL_KERNEL_LAUNCH_POINTERS_##type( name, type, __VA_ARGS__ )

#define CTRL_KERNEL_LAUNCH_POINTERS_GENERIC( name, type, ... )

#ifdef _CTRL_ARCH_FPGA_
	#define CTRL_KERNEL_LAUNCH_POINTERS_FPGA( name, type, ... ) \
		static cl_kernel *p_kernel_##type##_##name = NULL; \
		static cl_program *p_program_##type##_##name = NULL; \
		static char *p_kernel_raw_##type##_##name = NULL; \
		static const char *p_ctrl_kernel_name_##type##_##name = CTRL_KERNEL_STRINGIFY( CTRL_KERNEL_##type##_##name ) ;
#endif

#ifdef _CTRL_ARCH_OPENCL_GPU_
	#define CTRL_KERNEL_LAUNCH_POINTERS_OPENCL_GPU( name, type, ... )
#endif

#ifdef _CTRL_ARCH_CPU_
	#define CTRL_KERNEL_LAUNCH_POINTERS_CPU( name, type, ... )
#endif

#ifdef _CTRL_ARCH_CUDA_
	#define CTRL_KERNEL_LAUNCH_POINTERS_CUDA( name, type, ... )
#endif

#ifdef _CTRL_ARCH_XPHI_
	#define CTRL_KERNEL_LAUNCH_POINTERS_XPHI( name, type, ... )
#endif

#define _STRINGIFY(x) #x
#define STRINGIFY(x) _STRINGIFY(x)

// /* @author: Gabriel Rodriguez-Canal
//    @brief: Adds _profiling to the kernel name when profiling mode is chosen */
// #ifdef FPGA_PROFILING
// 	#define KERNEL_NAME(name) STRINGIFY(name##_profiling)
// #elif FPGA_EMULATION
// 	#define KERNEL_NAME(name) STRINGIFY(name##_emu)
// #else
// 	#define KERNEL_NAME(name) STRINGIFY(name)
// #endif // To remove

#define CTRL_INIT_KERNEL( name, type, ... ) \
	CTRL_INIT_KERNEL_##type( name, type, __VA_ARGS__ )

#ifdef _CTRL_ARCH_FPGA_
	#define CTRL_INIT_KERNEL_FPGA( name, type, ... ) \
		extern cl_kernel *p_kernel_##type##_##name; \
		extern cl_program *p_program_##type##_##name; \
		extern char *p_kernel_raw_##type##_##name; \
		extern const char *p_ctrl_kernel_name_##type##_##name; \
		extern int kernel_model; \
		extern size_t local_size_FPGA_##name [3]; /* \
		\
		__attribute__((constructor)) static void Ctrl_InitKernel_##type##_##name() { \
			p_kernel_##type##_##name = (cl_kernel *)malloc(sizeof(cl_kernel)); \
			p_program_##type##_##name = (cl_program *)malloc(sizeof(cl_program)); \
			FILE *binary_file; \
			static char *kernel_path = NULL; \
			kernel_path = (char *)malloc(CTRL_KERNEL_PATH_LENGTH * sizeof(char)); \
			kernel_path[0] = '\0'; \
			strcat(kernel_path, STRINGIFY(CTRL_KERNEL_PATH)); \
			strcat(kernel_path, CTRL_KERNEL_STRINGIFY(name)); \
			strcat(kernel_path, "/"); \
			strcat(kernel_path, KERNEL_NAME(name)); \
			strcat(kernel_path, "_Ctrl.aocx"); \
			kernel_path = "/home/u46403/Controllers/controllers/kernels/FPGA/Ctrl/Mult/Mult_SIMD-64x64x1_emu_Ctrl.aocx"; \
			printf("%s\n", kernel_path); \
			if(!(binary_file = fopen(kernel_path, "rb"))) { \
				printf("Kernel file not found.\n"); \
				exit(ERR_NOT_FOUND); \
			} \
			fseek(binary_file, 0, SEEK_END); \
			size_t binary_length = ftell(binary_file); \
			unsigned char *binary_str = (unsigned char*)malloc(binary_length * sizeof(unsigned char)); \
			rewind(binary_file); \
			if(!(fread(binary_str, binary_length, 1, binary_file))) { \
				printf("Error reading kernel file\n"); \
				exit(ERR_READ); \
			} \
			\
			struct kernel_params* curr_k_par = &kernel_params; \
			while(curr_k_par->next != NULL) \
				curr_k_par = curr_k_par->next; \
			\
			curr_k_par->kernel = p_kernel_##type##_##name; \
			curr_k_par->program = p_program_##type##_##name; \
			curr_k_par->kernel_name = p_ctrl_kernel_name_##type##_##name; \
			curr_k_par->binary_str = binary_str; \
			curr_k_par->binary_length = binary_length; \
			static struct kernel_params new_kernel_params; \
			curr_k_par->next = &new_kernel_params; \
		} */
#endif

#ifdef _CTRL_ARCH_CUDA_
	#define CTRL_INIT_KERNEL_CUDA( name, type, ... ) \
		int CTRL_KERNEL_CUDA_CHAR_type_##name = CTRL_KERNEL_CUDA_DYNAMIC; \
		dim3 CTRL_KERNEL_CUDA_CHAR_block_##name [0];

	#define CTRL_INIT_KERNEL_GENERIC( name, type, ... ) \
			int CTRL_KERNEL_CUDA_CHAR_type_##name = CTRL_KERNEL_CUDA_DYNAMIC; \
			dim3 CTRL_KERNEL_CUDA_CHAR_block_##name [0];
#endif

#ifdef _CTRL_ARCH_OPENCL_GPU_
	#define CTRL_INIT_KERNEL_OPENCL_GPU( name, type, ... ) \
		size_t local_size_OPENCL_GPU_##name [3];

	#define CTRL_INIT_KERNEL_GENERIC( name, type, ... ) \
			size_t local_size_OPENCL_GPU_##name [3];
#endif

#ifdef _CTRL_ARCH_XPHI_
	#define CTRL_INIT_KERNEL_XPHI( name, type, ... )
#endif

#define CTRL_HOST_TASK( name, ... ) \
    void Ctrl_Host_Task_##name(CTRL_KERNEL_EXTRACT_ARGS( __VA_ARGS__ )) { \
        CTRL_KERNEL_EXTRACT_KERNEL_NO_STR( __VA_ARGS__ ) \
    }
		

#define CTRL_HOST_TASK_PROTO( name, n_args, ... ) \
	\
	void Ctrl_HostWrapper_##name( void * args_list ) { \
		Ctrl_Host_Task_##name( CTRL_KERNEL_ARG_LIST_ACCESS_TILE( args_list, n_args, __VA_ARGS__ ) ); \
	} \
	\
	Ctrl_Task Ctrl_HostTaskCreate_##name(CTRL_KERNEL_TYPED_POINTER( NULL, n_args, __VA_ARGS__ )) { \
		Ctrl_Task task = CTRL_TASK_NULL; \
		task.task_type = CTRL_TASK_TYPE_HOST; \
		task.n_arguments = n_args; \
		task.pfn_hostTask_wrapper = Ctrl_HostWrapper_##name; \
		CTRL_KERNEL_ARG_LIST_CREATE_TILE( task.p_arguments, n_args, __VA_ARGS__ ); \
		CTRL_KERNEL_ROLES( task.p_roles, n_args, __VA_ARGS__ ); \
		CTRL_KERNEL_POINTERS( task.pp_pointers, n_args, __VA_ARGS__ ); \
		CTRL_KERNEL_DISPLACEMENTS_TILES( task.p_displacements, n_args, __VA_ARGS__ ); \
		task.device_id = -1; \
		return task; \
	}
		
#endif // _CTRL_KERNEL_PROTO_H_
