/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include "Architectures/OpenCL/Ctrl_OpenCL_Gpu.h"

void Ctrl_OpenCLGpu_EvalTaskKernelLaunch(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskHostTaskLaunch(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskGlobalSync(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskCreateTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskAllocTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskFreeTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskMoveTo(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskMoveToInner(Ctrl_OpenCLGpu *p_ctrl, HitTile *p_tile);
void Ctrl_OpenCLGpu_EvalTaskMoveFrom(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);
void Ctrl_OpenCLGpu_EvalTaskMoveFromInner(Ctrl_OpenCLGpu *p_ctrl, HitTile *p_tile);
void Ctrl_OpenCLGpu_EvalTaskWaitTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task);

void Ctrl_OpenCLGpu_Sync(Ctrl_OpenCLGpu *p_ctrl);

void Ctrl_OpenCLGpu_Create(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Policy policy, int device, int platform) {
	cl_int err;

	p_ctrl->policy = policy;

	cl_platform_id *p_platform_ids = (cl_platform_id *)malloc((platform + 1) * sizeof(cl_platform_id));
	OPENCL_ASSERT_OP( clGetPlatformIDs(platform + 1, p_platform_ids, NULL) );
	p_ctrl->platform_id = p_platform_ids[platform];
	free(p_platform_ids);

	cl_device_id *p_device_ids = (cl_device_id *)malloc((device + 1) * sizeof(cl_device_id));
	OPENCL_ASSERT_OP( clGetDeviceIDs(p_ctrl->platform_id, CL_DEVICE_TYPE_GPU, device + 1, p_device_ids, NULL) );
	p_ctrl->device_id = p_device_ids[device];
	free(p_device_ids);

	cl_context_properties context_properties[] = {CL_CONTEXT_PLATFORM, (cl_context_properties)p_ctrl->platform_id, 0};
	p_ctrl->context = clCreateContext(context_properties, 1, &(p_ctrl->device_id), NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	
	p_ctrl->queue_properties = 0;
	#ifdef _CTRL_OPENCL_GPU_PROFILING_
		p_ctrl->queue_properties |= CL_QUEUE_PROFILING_ENABLE;
	#endif

	p_ctrl->p_tile_list_head = NULL;
	p_ctrl->p_tile_list_tail = NULL;

	p_ctrl->queue = clCreateCommandQueue(p_ctrl->context, p_ctrl->device_id, p_ctrl->queue_properties, &err);
	OPENCL_ASSERT_ERROR( err );

	p_ctrl->default_event = clCreateUserEvent(p_ctrl->context, &err);
	OPENCL_ASSERT_ERROR( err );
	OPENCL_ASSERT_OP( clSetUserEventStatus(p_ctrl->default_event, CL_COMPLETE) );

	p_ctrl->last_kernel_event = p_ctrl->default_event;
	OPENCL_ASSERT_OP( clRetainEvent(p_ctrl->last_kernel_event) );
	p_ctrl->last_host_task_event = p_ctrl->default_event;
	OPENCL_ASSERT_OP( clRetainEvent(p_ctrl->last_host_task_event) );
	
	p_ctrl->p_resources_list = NULL;

	#ifdef _CTRL_OPENCL_GPU_PROFILING_
		p_ctrl->platform = platform;
		p_ctrl->device = device;
		
		p_ctrl->profiling_info_start = CL_PROFILING_COMMAND_START;
		p_ctrl->profiling_info_end = CL_PROFILING_COMMAND_END;

		p_ctrl->profiling_read_events = (cl_event *)malloc(_OPENCL_GPU_PROFILING_N_READ_TASKS_ * sizeof(cl_event));
		p_ctrl->profiling_write_events = (cl_event *)malloc(_OPENCL_GPU_PROFILING_N_WRITE_TASKS_ * sizeof(cl_event));
		p_ctrl->profiling_kernel_events = (cl_event *)malloc(_OPENCL_GPU_PROFILING_N_KERNEL_TASKS_ * sizeof(cl_event));

		p_ctrl->i_read_task = 0;
		p_ctrl->i_write_task = 0;
		p_ctrl->i_kernel_task = 0;

		p_ctrl->profiling_total = 0;
		p_ctrl->profiling_sum = 0;
		p_ctrl->profiling_offloading = 0;
		p_ctrl->profiling_read = 0;
		p_ctrl->profiling_write = 0;
		p_ctrl->profiling_kernel = 0;

		p_ctrl->profiling_start = 0;
		p_ctrl->profiling_end = 0;

		p_ctrl->has_first_profiling_event = false;

		p_ctrl->last_profiling_event = p_ctrl->default_event;
		OPENCL_ASSERT_OP( clRetainEvent(p_ctrl->last_profiling_event) );
		
		#ifdef _CTRL_OPENCL_GPU_PROFILING_VERBOSE_
			p_ctrl->profiling_visual_events = (visual_event *)malloc(
					(_OPENCL_GPU_PROFILING_N_READ_TASKS_ + 
					_OPENCL_GPU_PROFILING_N_WRITE_TASKS_ + 
					_OPENCL_GPU_PROFILING_N_KERNEL_TASKS_) * 
					sizeof(visual_event)
			);
			p_ctrl->i_visual_task = 0;
		#endif
	#endif // _CTRL_OPENCL_GPU_PROFILING_
}


void Ctrl_OpenCLGpu_Destroy(Ctrl_OpenCLGpu *p_ctrl) {
	#ifdef _CTRL_OPENCL_GPU_PROFILING_
		#ifdef _OPENCL_GPU_TEST_OUTPUT_
			printf("%d %d\n", p_ctrl->platform, p_ctrl->device);
			printf("%d %d %d %d %d\n", p_ctrl->i_read_task + p_ctrl->i_write_task + p_ctrl->i_kernel_task,
					p_ctrl->i_kernel_task, p_ctrl->i_read_task + p_ctrl->i_write_task, p_ctrl->i_read_task, p_ctrl->i_write_task);
		#else
			printf("\n ----------------------- ARGS ----------------------- \n");
			printf("\n PLATFORM: %d", p_ctrl->platform);
			printf("\n DEVICE: %d", p_ctrl->device);
			printf("\n TOTAL TASKS: %d", p_ctrl->i_read_task + p_ctrl->i_write_task + p_ctrl->i_kernel_task);
			printf("\n KERNEL TASKS: %d", p_ctrl->i_kernel_task);
			printf("\n OFFLOADING TASKS: %d", p_ctrl->i_read_task + p_ctrl->i_write_task);
			printf("\n READ TASKS: %d", p_ctrl->i_read_task);
			printf("\n WRITE TASKS: %d", p_ctrl->i_write_task);
			if (p_ctrl->policy == CTRL_POLICY_ASYNC) {
				printf("\n POLICY ASYNC");
			} else {
				printf("\n POLICY SYNC");
			}
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _OPENCL_GPU_TEST_OUTPUT_

		#ifdef _CTRL_OPENCL_GPU_PROFILING_VERBOSE_
			FILE *f = fopen("visual_profiler_info.txt", "w");

			for (int i = 0; i < p_ctrl->i_visual_task; i++) {
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_visual_events[i].event, p_ctrl->profiling_info_start, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_start)), NULL) );
				OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_visual_events[i].event, p_ctrl->profiling_info_end, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_end)), NULL) );

				fprintf(f, "%d; %d; %lu; %lu\n", p_ctrl->profiling_visual_events[i].queue, p_ctrl->profiling_visual_events[i].op, p_ctrl->profiling_start, p_ctrl->profiling_end);
				OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->profiling_visual_events[i].event) );
			}

			fclose(f);
			free(p_ctrl->profiling_visual_events);
		#endif

		for (int i = 0; i < p_ctrl->i_read_task; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_read_events[i], p_ctrl->profiling_info_start, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_start)), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_read_events[i], p_ctrl->profiling_info_end, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_end)), NULL) );

			p_ctrl->profiling_sum += p_ctrl->profiling_end - p_ctrl->profiling_start;
			p_ctrl->profiling_read += p_ctrl->profiling_end - p_ctrl->profiling_start;
			p_ctrl->profiling_offloading += p_ctrl->profiling_end - p_ctrl->profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->profiling_read_events[i]) );
		}
		free(p_ctrl->profiling_read_events);

		for (int i = 0; i < p_ctrl->i_write_task; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_write_events[i], p_ctrl->profiling_info_start, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_start)), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_write_events[i], p_ctrl->profiling_info_end, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_end)), NULL) );

			p_ctrl->profiling_sum += p_ctrl->profiling_end - p_ctrl->profiling_start;
			p_ctrl->profiling_write += p_ctrl->profiling_end - p_ctrl->profiling_start;
			p_ctrl->profiling_offloading += p_ctrl->profiling_end - p_ctrl->profiling_start;
			
			OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->profiling_write_events[i]) );
		}
		free(p_ctrl->profiling_write_events);

		for (int i = 0; i < p_ctrl->i_kernel_task; i++) {
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_kernel_events[i], p_ctrl->profiling_info_start, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_start)), NULL) );
			OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->profiling_kernel_events[i], p_ctrl->profiling_info_end, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_end)), NULL) );
			
			p_ctrl->profiling_sum += p_ctrl->profiling_end - p_ctrl->profiling_start;
			p_ctrl->profiling_kernel += p_ctrl->profiling_end - p_ctrl->profiling_start;

			OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->profiling_kernel_events[i]) );
		}
		free(p_ctrl->profiling_kernel_events);


		OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->first_profiling_event, p_ctrl->profiling_info_start, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_start)), NULL) );
		OPENCL_ASSERT_OP( clGetEventProfilingInfo(p_ctrl->last_profiling_event, p_ctrl->profiling_info_end, sizeof(cl_ulong), (void *)(&(p_ctrl->profiling_end)), NULL) );

		p_ctrl->profiling_total = p_ctrl->profiling_end - p_ctrl->profiling_start;

		OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->first_profiling_event) );
		OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->last_profiling_event) );

		#ifdef _OPENCL_GPU_TEST_OUTPUT_
			printf("%lf %lf %lf %lf %lf\n", p_ctrl->profiling_total / 1000000.0, p_ctrl->profiling_sum / 1000000.0, p_ctrl->profiling_kernel / 1000000.0, p_ctrl->profiling_offloading / 1000000.0, p_ctrl->profiling_read / 1000000.0, p_ctrl->profiling_write / 1000000.0);
		#else
			printf("\n -------------------- PROFILING --------------------- \n");
			printf("\n Total Time: %.3lf", p_ctrl->profiling_total / 1000000.0);
			printf("\n Acumulate Sum Time: %.3lf", p_ctrl->profiling_sum / 1000000.0);
			printf("\n Kernel Time: %.3lf", p_ctrl->profiling_kernel / 1000000.0);
			printf("\n Offloading Time: %.3lf", p_ctrl->profiling_offloading / 1000000.0);
			printf("\n Read Time: %.3lf", p_ctrl->profiling_read / 1000000.0);
			printf("\n Write Time: %.3lf", p_ctrl->profiling_write / 1000000.0);
			printf("\n\n ---------------------------------------------------- \n");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_TEST_MODE_
	#endif // _CTRL_OPENCL_GPU_PROFILING_

	OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->last_kernel_event) );
	OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->last_host_task_event) );
	OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->default_event) );

	Ctrl_OpenCL_Res *p_aux = NULL;
	Ctrl_OpenCL_Res *p_res = p_ctrl->p_resources_list;
	while(p_res != NULL) {
		p_aux = p_res;
		p_res = p_res->p_next;
		OPENCL_ASSERT_OP( clReleaseKernel(*(p_aux->p_kernel)) );
		OPENCL_ASSERT_OP( clReleaseProgram(*(p_aux->p_program)) );
		free(p_aux->p_kernel);
		free(p_aux->p_program);
		free(p_aux->p_kernel_raw);
		free(p_aux);
	}

	p_ctrl->p_tile_list_head = NULL;
	p_ctrl->p_tile_list_tail = NULL;

	OPENCL_ASSERT_OP( clReleaseCommandQueue(p_ctrl->queue) );
	OPENCL_ASSERT_OP( clReleaseContext(p_ctrl->context) );
}


void Ctrl_OpenCLGpu_Sync(Ctrl_OpenCLGpu *p_ctrl) {
	omp_set_lock(p_ctrl->p_lock_first_ctrl);
	omp_unset_lock(p_ctrl->p_lock_ctrl);
	omp_set_lock(p_ctrl->p_lock_host);
	omp_unset_lock(p_ctrl->p_lock_first_ctrl);

	omp_set_lock(p_ctrl->p_lock_first_host);
	omp_unset_lock(p_ctrl->p_lock_host);
	omp_set_lock(p_ctrl->p_lock_ctrl);
	omp_unset_lock(p_ctrl->p_lock_first_host);
}


void Ctrl_OpenCLGpu_EvalTask(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	switch ( p_task->task_type ) {
		case CTRL_TASK_TYPE_KERNEL: 
			Ctrl_OpenCLGpu_EvalTaskKernelLaunch(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_HOST: 
			Ctrl_OpenCLGpu_EvalTaskHostTaskLaunch(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_GLOBALSYNC:
			Ctrl_OpenCLGpu_EvalTaskGlobalSync(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_ALLOCTILE:
			Ctrl_OpenCLGpu_EvalTaskAllocTile(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_FREETILE:
			Ctrl_OpenCLGpu_EvalTaskFreeTile(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_MOVETO:
			Ctrl_OpenCLGpu_EvalTaskMoveTo(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_MOVEFROM:
			Ctrl_OpenCLGpu_EvalTaskMoveFrom(p_ctrl, p_task);
			break;
		case CTRL_TASK_TYPE_WAITTILE:
			Ctrl_OpenCLGpu_EvalTaskWaitTile(p_ctrl, p_task);
			break;
	}
}


void Ctrl_OpenCLGpu_EvalTaskGlobalSync(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	OPENCL_ASSERT_OP( clFinish(p_ctrl->queue) );
	for (Ctrl_OpenCL_Tile_List *p_aux = p_ctrl->p_tile_list_head; p_aux != NULL; p_aux = p_aux->p_next) {
		OPENCL_ASSERT_OP( clFinish(p_aux->p_tile_ext->queue) );
	}
	OPENCL_ASSERT_OP( clWaitForEvents(1, &(p_ctrl->last_host_task_event)) );

	Ctrl_OpenCLGpu_Sync(p_ctrl);
}


void Ctrl_OpenCLGpu_EvalTaskKernelLaunch(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	int n_event_wait = 0;
	cl_event p_event_wait_list[60];

	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			KHitTile *p_ktile = (KHitTile *)((uint8_t *)p_task->p_arguments + p_task->p_displacements[i]);
			Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_OUT) {
				if (p_tile_data->last_update == CTRL_TILE_LAST_UPDATE_HOST) {
					Ctrl_OpenCLGpu_EvalTaskMoveToInner(p_ctrl, p_tile);
				}
			}

			p_ktile->data = &(p_tile_data->device_data);

			p_event_wait_list[n_event_wait] = p_tile_data->kernel_last_write_event;
			n_event_wait++;

			p_event_wait_list[n_event_wait] = p_tile_data->offloading_last_write_event;
			n_event_wait++;

			if (p_task->p_roles[i] != KERNEL_IN) {
				p_event_wait_list[n_event_wait] = p_tile_data->kernel_last_read_event;
				n_event_wait++;
				p_event_wait_list[n_event_wait] = p_tile_data->offloading_last_read_event;
				n_event_wait++;
			}
		}
	}

	cl_event event_aux = p_ctrl->last_kernel_event;

	Ctrl_Request request;

	request.opencl.context = &(p_ctrl->context);
	request.opencl.queue = &(p_ctrl->queue);
	request.opencl.device_id = &(p_ctrl->device_id);
	request.opencl.p_last_kernel_event = &(p_ctrl->last_kernel_event);
	request.opencl.p_event_wait_list = p_event_wait_list;
	request.opencl.n_event_wait = n_event_wait;
	request.opencl.n_arguments = p_task->n_arguments;
	request.opencl.p_roles = p_task->p_roles;
	request.opencl.p_displacements = p_task->p_displacements;
	request.opencl.p_res_list_head = p_ctrl->p_resources_list;
	
	p_task->pfn_kernel_wrapper(request, CTRL_TYPE_OPENCL_GPU, p_task->device_id, p_task->threads, p_task->p_arguments);

	OPENCL_ASSERT_OP( clReleaseEvent(event_aux) );

	OPENCL_PROFILE_VISUAL( CTRL_OPENCL_PROFILE_QUEUE_MAIN, p_ctrl->last_kernel_event, CTRL_OPENCL_PROFILE_OP_KERNEL );
	OPENCL_PROFILE_KERNEL( p_ctrl->last_kernel_event );

	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_IN) {
				p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_DEV;

				OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->kernel_last_write_event) );
				p_tile_data->kernel_last_write_event = p_ctrl->last_kernel_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->kernel_last_write_event) );
			}

			if (p_task->p_roles[i] != KERNEL_OUT) {
				OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->kernel_last_read_event) );
				p_tile_data->kernel_last_read_event = p_ctrl->last_kernel_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->kernel_last_read_event) );
			}
		}
	}
}


void Ctrl_OpenCLGpu_EvalTaskHostTaskLaunch(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	cl_int err;

	int n_event_wait = 0;
	cl_event *p_event_wait_list = (cl_event *)malloc(sizeof(cl_event) * 60);

	int n_event_release = 0;
	cl_event *p_event_release_list = (cl_event *)malloc(sizeof(cl_event) * 60);
	
	cl_event host_task_event = clCreateUserEvent(p_ctrl->context, &err);
	OPENCL_ASSERT_ERROR( err );

	OPENCL_ASSERT_OP( clReleaseEvent(p_ctrl->last_host_task_event) );
	p_ctrl->last_host_task_event = host_task_event;
	OPENCL_ASSERT_OP( clRetainEvent(p_ctrl->last_host_task_event) );
	
	
	for (int i = 0; i < p_task->n_arguments; i++) {
		if (p_task->p_roles[i] != KERNEL_INVAL) {
			HitTile *p_tile = (HitTile *)(p_task->pp_pointers[i]);
			Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

			if (p_task->p_roles[i] != KERNEL_OUT) {
				if (p_tile_data->last_update == CTRL_TILE_LAST_UPDATE_DEV) {
					Ctrl_OpenCLGpu_EvalTaskMoveFromInner(p_ctrl, p_tile);
				}
			}

			p_event_wait_list[n_event_wait] = p_tile_data->offloading_last_read_event;
			OPENCL_ASSERT_OP( clRetainEvent(p_event_wait_list[n_event_wait]) );
			n_event_wait++;

			p_event_wait_list[n_event_wait] = p_tile_data->host_last_write_event;
			OPENCL_ASSERT_OP( clRetainEvent(p_event_wait_list[n_event_wait]) );
			n_event_wait++;

			if (p_task->p_roles[i] != KERNEL_IN) {
				p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST;
				p_event_wait_list[n_event_wait] = p_tile_data->offloading_last_write_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_event_wait_list[n_event_wait]) );
				n_event_wait++;
				p_event_wait_list[n_event_wait] = p_tile_data->host_last_read_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_event_wait_list[n_event_wait]) );
				n_event_wait++;

				p_event_release_list[n_event_release] = p_tile_data->host_last_write_event;
				n_event_release++;
				p_tile_data->host_last_write_event = host_task_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->host_last_write_event) );
			}

			if (p_task->p_roles[i] != KERNEL_OUT) {
				p_event_release_list[n_event_release] = p_tile_data->host_last_read_event;
				n_event_release++;
				p_tile_data->host_last_read_event = host_task_event;
				OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->host_last_read_event) );
			}
		}
	}

	if (p_ctrl->policy != CTRL_POLICY_ASYNC) {
		OPENCL_ASSERT_OP( clEnqueueMarkerWithWaitList(p_ctrl->queue, 0, NULL, &p_event_wait_list[n_event_wait]) );
		OPENCL_ASSERT_OP( clRetainEvent(p_event_wait_list[n_event_wait]) );
		n_event_wait++;
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		if (n_event_wait > 0) {
			OPENCL_ASSERT_OP( clWaitForEvents(n_event_wait, p_event_wait_list) );
		}
		
		#pragma omp task
		{
			p_task->pfn_hostTask_wrapper(p_task->p_arguments);
			OPENCL_ASSERT_OP( clSetUserEventStatus(host_task_event, CL_COMPLETE) );
		}
		OPENCL_ASSERT_OP( clWaitForEvents(1, &host_task_event) );

		for (int i = 0; i < n_event_wait; i++) {
			OPENCL_ASSERT_OP( clReleaseEvent(p_event_wait_list[i]) );
		}

		for (int i = 0; i < n_event_release; i++) {
			OPENCL_ASSERT_OP( clReleaseEvent(p_event_release_list[i]) );
		}

		Ctrl_OpenCLGpu_Sync(p_ctrl);
		
	} else {
		#pragma omp task depend(inout: p_ctrl)
		{
			if (n_event_wait > 0) {
				OPENCL_ASSERT_OP( clWaitForEvents(n_event_wait, p_event_wait_list) );
			}

			p_task->pfn_hostTask_wrapper(p_task->p_arguments);
			OPENCL_ASSERT_OP( clSetUserEventStatus(host_task_event, CL_COMPLETE) );
			
			for (int i = 0; i < n_event_wait; i++) {
				OPENCL_ASSERT_OP( clReleaseEvent(p_event_wait_list[i]) );
			}

			for (int i = 0; i < n_event_release; i++) {
				OPENCL_ASSERT_OP( clReleaseEvent(p_event_release_list[i]) );
			}
		}
	}
}


void Ctrl_OpenCLGpu_EvalTaskCreateTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)malloc(sizeof(Ctrl_OpenCL_Tile));

	p_tile_data->p_ctrl = p_ctrl;

	p_tile_data->kernel_last_read_event = p_ctrl->default_event;
	p_tile_data->kernel_last_write_event = p_ctrl->default_event;

	p_tile_data->offloading_last_read_event = p_ctrl->default_event;
	p_tile_data->offloading_last_write_event = p_ctrl->default_event;

	p_tile_data->host_last_read_event = p_ctrl->default_event;
	p_tile_data->host_last_write_event = p_ctrl->default_event;

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;

	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->kernel_last_read_event) );
	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->kernel_last_write_event) );

	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->offloading_last_read_event) );
	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->offloading_last_write_event) );

	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->host_last_read_event) );
	OPENCL_ASSERT_OP( clRetainEvent(p_tile_data->host_last_write_event) );

	p_tile->ext = (void *)p_tile_data;
}


void Ctrl_OpenCLGpu_EvalTaskAllocTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	Ctrl_OpenCLGpu_EvalTaskCreateTile(p_ctrl, p_task);

	cl_int err;

	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	Ctrl_OpenCL_Tile_List *p_list_node = (Ctrl_OpenCL_Tile_List *)malloc(sizeof(Ctrl_OpenCL_Tile_List));
	
	p_list_node->p_prev = NULL;
	p_list_node->p_next = NULL;
	p_list_node->p_tile_ext = p_tile_data;
	p_tile_data->p_tile_elem = p_list_node;

	if (p_ctrl->p_tile_list_tail != NULL){
		p_ctrl->p_tile_list_tail->p_next = p_list_node;
		p_list_node->p_prev = p_ctrl->p_tile_list_tail;
		p_ctrl->p_tile_list_tail = p_list_node;
	} else {
		p_ctrl->p_tile_list_head = p_ctrl->p_tile_list_tail = p_list_node;
	}

	p_tile_data->queue = clCreateCommandQueue(p_ctrl->context, p_ctrl->device_id, p_ctrl->queue_properties, &err);
	OPENCL_ASSERT_ERROR( err );

	p_tile_data->device_data = clCreateBuffer(p_ctrl->context, CL_MEM_READ_WRITE, ((size_t)(p_tile->acumCard)) * (p_tile->baseExtent), NULL, &err);
	OPENCL_ASSERT_ERROR( err );

	p_tile_data->pinned_data = clCreateBuffer(p_ctrl->context, CL_MEM_ALLOC_HOST_PTR, (size_t)p_tile->acumCard * p_tile->baseExtent, NULL, &err);
	OPENCL_ASSERT_ERROR( err );


	p_tile->data = clEnqueueMapBuffer(p_tile_data->queue, p_tile_data->pinned_data, CL_TRUE, CL_MAP_READ | CL_MAP_WRITE, 0, (size_t)p_tile->acumCard * p_tile->baseExtent, 0, NULL, NULL, &err);
	OPENCL_ASSERT_ERROR( err );
	OPENCL_ASSERT_OP( clFlush(p_tile_data->queue) );

	p_tile->memPtr = p_tile->data;

	if ( (p_tile->memStatus == HIT_MS_NOT_OWNER) || (p_tile->memStatus == HIT_MS_NOMEM) ) {
		for (int i = 0; i < hit_shapeDims(p_tile->shape); i++) {
			p_tile->qstride[i] = 1;
		}
		
		int cardinality = 1;
		p_tile->origAcumCard[hit_shapeDims(p_tile->shape)] = 1;

		for (int i = (hit_shapeDims(p_tile->shape) - 1); i >= 0; i--) {
			cardinality = cardinality * p_tile->card[i];
			p_tile->origAcumCard[i] = cardinality;
		}
		p_tile->acumCard = p_tile->origAcumCard[0];
	}

	p_tile->memStatus = HIT_MS_OWNER;
	
	Ctrl_OpenCLGpu_Sync(p_ctrl);
}


void Ctrl_OpenCLGpu_EvalTaskFreeTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	
	OPENCL_ASSERT_OP( clEnqueueUnmapMemObject(p_tile_data->queue, p_tile_data->pinned_data, p_tile->data, 0, NULL, NULL) );
	OPENCL_ASSERT_OP( clFlush(p_tile_data->queue) );
	OPENCL_ASSERT_OP( clFinish(p_tile_data->queue) );
	OPENCL_ASSERT_OP( clReleaseMemObject(p_tile_data->pinned_data) );

	if (p_tile->memStatus == HIT_MS_OWNER) {
		p_tile->memPtr = NULL;
		p_tile->data = NULL;
		p_tile->memStatus = HIT_MS_NOMEM;
	}
	

	OPENCL_ASSERT_OP( clReleaseMemObject(p_tile_data->device_data) );
	
	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->kernel_last_read_event) );
	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->kernel_last_write_event) );

	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->offloading_last_read_event) );
	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->offloading_last_write_event) );

	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->host_last_read_event) );
	OPENCL_ASSERT_OP( clReleaseEvent(p_tile_data->host_last_write_event) );

	OPENCL_ASSERT_OP( clReleaseCommandQueue(p_tile_data->queue) );

	if (p_tile_data->p_tile_elem->p_prev != NULL) {
		if (p_tile_data->p_tile_elem->p_next == NULL) {
			p_tile_data->p_tile_elem->p_prev->p_next = NULL;
			p_ctrl->p_tile_list_tail = p_tile_data->p_tile_elem->p_prev;
		} else {
			p_tile_data->p_tile_elem->p_prev->p_next = p_tile_data->p_tile_elem->p_next;
		}
	}

	if (p_tile_data->p_tile_elem->p_next != NULL) {
		if (p_tile_data->p_tile_elem->p_prev == NULL) {
			p_tile_data->p_tile_elem->p_next->p_prev = NULL;
			p_ctrl->p_tile_list_head = p_tile_data->p_tile_elem->p_next;
		} else {
			p_tile_data->p_tile_elem->p_next->p_prev = p_tile_data->p_tile_elem->p_prev;
		}
	}

	p_tile->ext = NULL;
	p_tile_data->p_ctrl = NULL;
	
	p_tile_data->p_tile_elem->p_tile_ext = NULL;
	p_tile_data->p_tile_elem->p_next = NULL;
	p_tile_data->p_tile_elem->p_prev = NULL;

	free(p_tile_data->p_tile_elem);
	free(p_tile_data);

	Ctrl_OpenCLGpu_Sync(p_ctrl);
}


void Ctrl_OpenCLGpu_EvalTaskMoveTo(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	if (p_tile_data->last_update != CTRL_TILE_LAST_UPDATE_HOST_DEV) {
		Ctrl_OpenCLGpu_EvalTaskMoveToInner(p_ctrl, p_tile);
	}
	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		OPENCL_ASSERT_OP( clWaitForEvents(1, &(p_tile_data->offloading_last_write_event)) );
		Ctrl_OpenCLGpu_Sync(p_ctrl);
	}
}

void Ctrl_OpenCLGpu_EvalTaskMoveToInner(Ctrl_OpenCLGpu *p_ctrl, HitTile *p_tile) {
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);
	
	cl_event aux = p_tile_data->offloading_last_write_event;

	cl_event wait_list[4];
	wait_list[0] = p_tile_data->kernel_last_read_event;
	wait_list[1] = p_tile_data->kernel_last_write_event;
	wait_list[2] = p_tile_data->offloading_last_read_event;
	wait_list[3] = p_tile_data->host_last_write_event;

	OPENCL_ASSERT_OP( 
		clEnqueueWriteBuffer(
			p_tile_data->queue, 
			p_tile_data->device_data, 
			CL_FALSE, 
			0, 
			((size_t)(p_tile->acumCard)) * (p_tile->baseExtent), 
			p_tile->data,
			4,
			wait_list, 
			&(p_tile_data->offloading_last_write_event) 
		)
	);

	OPENCL_ASSERT_OP( clFlush(p_tile_data->queue) );
	OPENCL_ASSERT_OP( clReleaseEvent(aux) );

	OPENCL_PROFILE_VISUAL( 
			CTRL_OPENCL_PROFILE_QUEUE_WRITE, 
			p_tile_data->offloading_last_write_event, 
			CTRL_OPENCL_PROFILE_OP_WRITE 
	);
	OPENCL_PROFILE_WRITE( p_tile_data->offloading_last_write_event );

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;
}

void Ctrl_OpenCLGpu_EvalTaskMoveFrom(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	if (p_tile_data->last_update != CTRL_TILE_LAST_UPDATE_HOST_DEV) {
		Ctrl_OpenCLGpu_EvalTaskMoveFromInner(p_ctrl, p_tile);
	}

	if (p_ctrl->policy == CTRL_POLICY_SYNC) {
		OPENCL_ASSERT_OP( clWaitForEvents(1, &(p_tile_data->offloading_last_read_event)) );
		Ctrl_OpenCLGpu_Sync(p_ctrl);
	}
}

void Ctrl_OpenCLGpu_EvalTaskMoveFromInner(Ctrl_OpenCLGpu *p_ctrl, HitTile *p_tile) {
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	cl_event aux = p_tile_data->offloading_last_read_event;

	cl_event wait_list[4];
	wait_list[0] = p_tile_data->kernel_last_write_event;
	wait_list[1] = p_tile_data->offloading_last_write_event;
	wait_list[2] = p_tile_data->host_last_read_event;
	wait_list[3] = p_tile_data->host_last_write_event;

	OPENCL_ASSERT_OP( 
		clEnqueueReadBuffer(
			p_tile_data->queue, 
			p_tile_data->device_data, 
			CL_FALSE, 
			0, 
			((size_t)(p_tile->acumCard)) * (p_tile->baseExtent), 
			p_tile->data,
			4,
			wait_list,
			&(p_tile_data->offloading_last_read_event) 
		)
	);

	OPENCL_ASSERT_OP( clFlush(p_tile_data->queue) );
	OPENCL_ASSERT_OP( clReleaseEvent(aux) );

	OPENCL_PROFILE_VISUAL( 
			CTRL_OPENCL_PROFILE_QUEUE_READ, 
			p_tile_data->offloading_last_read_event, 
			CTRL_OPENCL_PROFILE_OP_READ 
	);
	OPENCL_PROFILE_READ( p_tile_data->offloading_last_read_event );

	p_tile_data->last_update = CTRL_TILE_LAST_UPDATE_HOST_DEV;
}

void Ctrl_OpenCLGpu_EvalTaskWaitTile(Ctrl_OpenCLGpu *p_ctrl, Ctrl_Task *p_task) {
	HitTile *p_tile = (HitTile *)(p_task->p_tile);
	Ctrl_OpenCL_Tile *p_tile_data = (Ctrl_OpenCL_Tile *)(p_tile->ext);

	cl_event wait_list[6]; 
	wait_list[0] = p_tile_data->kernel_last_read_event;
	wait_list[1] = p_tile_data->kernel_last_write_event;
	wait_list[2] = p_tile_data->offloading_last_read_event;
	wait_list[3] = p_tile_data->offloading_last_write_event;
	wait_list[4] = p_tile_data->host_last_read_event;
	wait_list[5] = p_tile_data->host_last_write_event;

	OPENCL_ASSERT_OP( clWaitForEvents(6, wait_list) );
	Ctrl_OpenCLGpu_Sync(p_ctrl);
}
