/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include "Core/Ctrl_Core.h"

//Ctrl Ctrl_global;

void Ctrl_AddTask(Ctrl *p_ctrl, Ctrl_TaskType type, HitTile *p_tile);
void Ctrl_ConsumeTaskQueue(Ctrl *p_ctrl);
void Ctrl_CreateInner(Ctrl *p_ctrl, Ctrl_Type type, Ctrl_Policy policy, int device, void *ext);

void Ctrl_Sync(Ctrl *p_ctrl);
void Ctrl_SyncIfSync(Ctrl *p_ctrl);
void Ctrl_Thread_EvalThread();

int Ctrl_Thread_init() {
    if (omp_get_thread_num() == 0) {
        return 0;
    } else if (omp_get_thread_num() == 1) {
        Ctrl_Thread_EvalThread();
    } else {
        #pragma omp barrier
        #pragma omp barrier
        #pragma omp barrier
        #pragma omp barrier
    }
    return 1;
}
 
void Ctrl_Thread_EvalThread() {
    #pragma omp barrier

    Ctrl ctrl = Ctrl_global;
    omp_set_lock(ctrl.p_lock_ctrl);

    #pragma omp barrier

    Ctrl_ConsumeTaskQueue(&ctrl);
    
    switch(ctrl.type){

        #ifdef _CTRL_ARCH_CPU_
            case CTRL_TYPE_CPU:
                Ctrl_Cpu_Destroy(&(ctrl.p_impl->cpu));
                break;
        #endif // _CTRL_ARCH_CPU_
                
        #ifdef _CTRL_ARCH_CUDA_
            case CTRL_TYPE_CUDA:
                Ctrl_Cuda_Destroy(&(ctrl.p_impl->cuda));
                break;
        #endif // _CTRL_ARCH_CUDA_

        #ifdef _CTRL_ARCH_OPENCL_GPU_
            case CTRL_TYPE_OPENCL_GPU:
                Ctrl_OpenCLGpu_Destroy(&(ctrl.p_impl->opencl_gpu));
                break;
        #endif // _CTRL_ARCH_OPENCL_GPU_

        #ifdef _CTRL_ARCH_FPGA_
            case CTRL_TYPE_FPGA:
                Ctrl_FPGA_Destroy(&(ctrl.p_impl->fpga));
                break;
        #endif // _CTRL_ARCH_OPENCL_GPU_

        #ifdef _CTRL_ARCH_XPHI_
            case CTRL_TYPE_XPHI:
                Ctrl_XPhi_Destroy(&(ctrl.p_impl->xphi));
                break;
        #endif // _CTRL_ARCH_XPHI_

    }

    #pragma omp barrier

    omp_unset_lock(ctrl.p_lock_ctrl);

    #pragma omp barrier
}


/* @author: Gabriel Rodriguez-Canal
    Controller creation functions. Interface macro in was set as variadic to enable different parameters to be passed to specific backends, 
    e.g. execution mode (profiling, emu...) to FPGA's
*/
/* #ifdef _CTRL_ARCH_FPGA_
void Ctrl_CreateInner_CTRL_TYPE_FPGA(Ctrl *p_ctrl, Ctrl_Type type, Ctrl_Policy policy, int device, int exec_mode) {
    printf("Hola desde createinner_fpga\n");
    Ctrl_FPGA_Create((&(p_ctrl->p_impl->fpga)), policy, device, exec_mode );
    p_ctrl->p_impl->fpga.p_lock_first_host = p_ctrl->p_lock_first_host;
    p_ctrl->p_impl->fpga.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
    p_ctrl->p_impl->fpga.p_lock_host = p_ctrl->p_lock_host;
    p_ctrl->p_impl->fpga.p_lock_ctrl = p_ctrl->p_lock_ctrl;
}
#endif */

void Ctrl_CreateWrapper(Ctrl_Type type, Ctrl *p_ctrl, Ctrl_Policy policy) {
    Ctrl_global.p_impl = (Ctrl_Impl *)malloc(sizeof(Ctrl_Impl));
    Ctrl_global.p_task_queue = (Ctrl_TaskQueue *)malloc(sizeof(Ctrl_TaskQueue));
    Ctrl_global.p_lock_first_host = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_first_ctrl = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_host = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_ctrl = (omp_lock_t *)malloc(sizeof(omp_lock_t));

    p_ctrl->type = type;

	if(policy != CTRL_POLICY_DEFAULT) {
		p_ctrl->policy = policy;
	} else {
		if ( getenv("CTRL_ASYNC_MODE") ) {
			if (atoi( getenv("CTRL_ASYNC_MODE") ) > 0) {
				p_ctrl->policy = CTRL_POLICY_ASYNC;
			} else {
				p_ctrl->policy = CTRL_POLICY_SYNC;
			}
		} else {
			p_ctrl->policy = CTRL_POLICY_SYNC;
		}
	}
	
    omp_init_lock(p_ctrl->p_lock_first_host);
    omp_init_lock(p_ctrl->p_lock_first_ctrl);
    omp_init_lock(p_ctrl->p_lock_host);
    omp_init_lock(p_ctrl->p_lock_ctrl);

	Ctrl_TaskQueue_Init(p_ctrl->p_task_queue);

    #pragma omp barrier

    omp_set_lock(p_ctrl->p_lock_host);

    #pragma omp barrier
}

#ifdef _CTRL_ARCH_FPGA_
	PCtrl Ctrl_CreateWrapper_CTRL_TYPE_FPGA(Ctrl_Policy policy, int device, int platform, int exec_mode) {
        Ctrl_CreateWrapper(CTRL_TYPE_FPGA, &Ctrl_global, policy);

        PCtrl p_ctrl = &Ctrl_global;
        Ctrl_FPGA_Create((&(p_ctrl->p_impl->fpga)), policy, device, platform, exec_mode );
        p_ctrl->p_impl->fpga.p_lock_first_host = p_ctrl->p_lock_first_host;
        p_ctrl->p_impl->fpga.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
        p_ctrl->p_impl->fpga.p_lock_host = p_ctrl->p_lock_host;
        p_ctrl->p_impl->fpga.p_lock_ctrl = p_ctrl->p_lock_ctrl;
        
        return &Ctrl_global;
    }
#endif
#ifdef _CTRL_ARCH_OPENCL_GPU_
    PCtrl Ctrl_CreateWrapper_CTRL_TYPE_OPENCL_GPU(Ctrl_Policy policy, int device, int platform) {
        Ctrl_CreateWrapper(CTRL_TYPE_OPENCL_GPU, &Ctrl_global, policy);

        PCtrl p_ctrl = &Ctrl_global;
        Ctrl_OpenCLGpu_Create((&(p_ctrl->p_impl->opencl_gpu)), policy, device, platform);
        p_ctrl->p_impl->opencl_gpu.p_lock_first_host = p_ctrl->p_lock_first_host;
        p_ctrl->p_impl->opencl_gpu.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
        p_ctrl->p_impl->opencl_gpu.p_lock_host = p_ctrl->p_lock_host;
        p_ctrl->p_impl->opencl_gpu.p_lock_ctrl = p_ctrl->p_lock_ctrl;

        return &Ctrl_global;
    }
#endif
#ifdef _CTRL_ARCH_CUDA_
	PCtrl Ctrl_CreateWrapper_CTRL_TYPE_CUDA(Ctrl_Policy policy, int device) {
        Ctrl_CreateWrapper(CTRL_TYPE_CUDA, &Ctrl_global, policy);

        PCtrl p_ctrl = &Ctrl_global;
        Ctrl_Cuda_Create((&(p_ctrl->p_impl->cuda)), policy, device);
        p_ctrl->p_impl->cuda.p_lock_first_host = p_ctrl->p_lock_first_host;
        p_ctrl->p_impl->cuda.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
        p_ctrl->p_impl->cuda.p_lock_host = p_ctrl->p_lock_host;
        p_ctrl->p_impl->cuda.p_lock_ctrl = p_ctrl->p_lock_ctrl;

        return &Ctrl_global;
    }
#endif

/* PCtrl Ctrl_CreateWrapper(Ctrl_Type type, Ctrl_Policy policy, int device, void * ext) {
    Ctrl_global.p_impl = (Ctrl_Impl *)malloc(sizeof(Ctrl_Impl));
    Ctrl_global.p_task_queue = (Ctrl_TaskQueue *)malloc(sizeof(Ctrl_TaskQueue));
    Ctrl_global.p_lock_first_host = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_first_ctrl = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_host = (omp_lock_t *)malloc(sizeof(omp_lock_t));
    Ctrl_global.p_lock_ctrl = (omp_lock_t *)malloc(sizeof(omp_lock_t));

    Ctrl_CreateInner(&Ctrl_global, type, policy, device, ext);
    
    return &Ctrl_global;
} */

// void Ctrl_CreateInner(Ctrl *p_ctrl, Ctrl_Type type, Ctrl_Policy policy, int device, void *ext) {

// 	p_ctrl->type = type;

// 	if(policy != CTRL_POLICY_DEFAULT) {
// 		p_ctrl->policy = policy;
// 	} else {
// 		if ( getenv("CTRL_ASYNC_MODE") ) {
// 			if (atoi( getenv("CTRL_ASYNC_MODE") ) > 0) {
// 				p_ctrl->policy = CTRL_POLICY_ASYNC;
// 			} else {
// 				p_ctrl->policy = CTRL_POLICY_SYNC;
// 			}
// 		} else {
// 			p_ctrl->policy = CTRL_POLICY_SYNC;
// 		}
// 	}
	
//     omp_init_lock(p_ctrl->p_lock_first_host);
//     omp_init_lock(p_ctrl->p_lock_first_ctrl);
//     omp_init_lock(p_ctrl->p_lock_host);
//     omp_init_lock(p_ctrl->p_lock_ctrl);

// 	Ctrl_TaskQueue_Init(p_ctrl->p_task_queue);
    
// 	switch(p_ctrl->type){

// 		// #ifdef _CTRL_ARCH_CPU_
// 		// 	case CTRL_TYPE_CPU:
// 		// 		Ctrl_Cpu_Create(&(p_ctrl->impl.cpu), device, policy);
// 		// 		break;
// 		// #endif // _CTRL_ARCH_CPU_
				
// 		#ifdef _CTRL_ARCH_CUDA_
// 			case CTRL_TYPE_CUDA:
// 				Ctrl_Cuda_Create((&(p_ctrl->p_impl->cuda)), policy, device);
//                 p_ctrl->p_impl->cuda.p_lock_first_host = p_ctrl->p_lock_first_host;
//                 p_ctrl->p_impl->cuda.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
//                 p_ctrl->p_impl->cuda.p_lock_host = p_ctrl->p_lock_host;
//                 p_ctrl->p_impl->cuda.p_lock_ctrl = p_ctrl->p_lock_ctrl;
// 				break;
// 		#endif // _CTRL_ARCH_CUDA_

// 		#ifdef _CTRL_ARCH_OPENCL_GPU_
// 			case CTRL_TYPE_OPENCL_GPU:
// 				Ctrl_OpenCLGpu_Create((&(p_ctrl->p_impl->opencl_gpu)), policy, device, (*((int *)(ext))) );
//                 p_ctrl->p_impl->opencl_gpu.p_lock_first_host = p_ctrl->p_lock_first_host;
//                 p_ctrl->p_impl->opencl_gpu.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
//                 p_ctrl->p_impl->opencl_gpu.p_lock_host = p_ctrl->p_lock_host;
//                 p_ctrl->p_impl->opencl_gpu.p_lock_ctrl = p_ctrl->p_lock_ctrl;
// 				break;
// 		#endif // _CTRL_ARCH_OPENCL_GPU_
        
//         #ifdef _CTRL_ARCH_FPGA_
// 			case CTRL_TYPE_FPGA:
// 				Ctrl_FPGA_Create((&(p_ctrl->p_impl->fpga)), policy, device, (*((int *)(ext))) );
//                 p_ctrl->p_impl->fpga.p_lock_first_host = p_ctrl->p_lock_first_host;
//                 p_ctrl->p_impl->fpga.p_lock_first_ctrl = p_ctrl->p_lock_first_ctrl;
//                 p_ctrl->p_impl->fpga.p_lock_host = p_ctrl->p_lock_host;
//                 p_ctrl->p_impl->fpga.p_lock_ctrl = p_ctrl->p_lock_ctrl;
// 				break;
// 		#endif // _CTRL_ARCH_OPENCL_GPU_
// 		// #ifdef _CTRL_ARCH_XPHI_
// 		// 	case CTRL_TYPE_XPHI:
// 		// 		Ctrl_XPhi_Create(&(p_ctrl->impl.xphi), device, policy);
// 		// 		p_ctrl->impl.xphi.sem_destroy = &(p_ctrl->sem_destroy);
// 		// 		p_ctrl->impl.xphi.sem_sync = &(p_ctrl->sem_sync);
// 		// 		break;
// 		// #endif // _CTRL_ARCH_XPHI_

//         default:
//             fprintf(stderr, "Ctrl_Create: Not supported architecture. Recompile Ctrl library with the proper support\n");
//             exit(EXIT_FAILURE);
// 	}

//     #pragma omp barrier

//     omp_set_lock(p_ctrl->p_lock_host);

//     #pragma omp barrier
// }

void Ctrl_ConsumeTaskQueue(Ctrl *p_ctrl) {
    #ifdef _CTRL_ARCH_CUDA_
        Ctrl_Cuda_NewThreadSetup(&(p_ctrl->p_impl->cuda));
    #endif // _CTRL_ARCH_CUDA_

    bool finish = false;

    while (finish == false) {
        Ctrl_Task *p_task = NULL;
        // printf("<-- pop task\n"); fflush(stdout);
        p_task = Ctrl_TaskQueue_Pop(p_ctrl->p_task_queue);
        // printf("<-- popped task %d\n", p_task->task_type); fflush(stdout);
        if (p_task->task_type == CTRL_TASK_TYPE_DESTROYCNTRL) {
            finish = true;
        } else {
            switch (p_ctrl->type) {
                #ifdef _CTRL_ARCH_CPU_
                    case CTRL_TYPE_CPU:
                        Ctrl_Cpu_EvalTask(&(p_ctrl->p_impl->cpu), p_task);
                        break;
                #endif // _CTRL_ARCH_CPU_

                #ifdef _CTRL_ARCH_CUDA_
                    case CTRL_TYPE_CUDA:
                        Ctrl_Cuda_EvalTask(&(p_ctrl->p_impl->cuda), p_task);
                        break;
                #endif // _CTRL_ARCH_CUDA_

                #ifdef _CTRL_ARCH_OPENCL_GPU_
                    case CTRL_TYPE_OPENCL_GPU:
                        Ctrl_OpenCLGpu_EvalTask(&(p_ctrl->p_impl->opencl_gpu), p_task);
                        break;
                #endif // _CTRL_ARCH_OPENCL_GPU_

                #ifdef _CTRL_ARCH_FPGA_
                    case CTRL_TYPE_FPGA:
                        Ctrl_FPGA_EvalTask(&(p_ctrl->p_impl->fpga), p_task);
                        break;
                #endif // _CTRL_ARCH_OPENCL_GPU_

                #ifdef _CTRL_ARCH_XPHI_
                    case CTRL_TYPE_XPHI:
                        Ctrl_XPhi_EvalTask(&(p_ctrl->p_impl->xphi), p_task);
                        break;
                #endif // _CTRL_ARCH_XPHI_
            }
            if (p_task->task_type != CTRL_TASK_TYPE_HOST) {
                Ctrl_TaskQueue_FreeTask(p_task);
            } 
        }
    }
}


void Ctrl_Destroy(Ctrl *p_ctrl) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_DESTROYCNTRL, NULL);
    #pragma omp barrier

    omp_unset_lock(p_ctrl->p_lock_host);

    #pragma omp barrier

    Ctrl_TaskQueue_Destroy(p_ctrl->p_task_queue);

    omp_destroy_lock(p_ctrl->p_lock_first_host);
    omp_destroy_lock(p_ctrl->p_lock_first_ctrl);
    omp_destroy_lock(p_ctrl->p_lock_host);
    omp_destroy_lock(p_ctrl->p_lock_ctrl);

    free(p_ctrl->p_impl);
    free(p_ctrl->p_task_queue);
    free(p_ctrl->p_lock_first_host);
    free(p_ctrl->p_lock_first_ctrl);
    free(p_ctrl->p_lock_host);
    free(p_ctrl->p_lock_ctrl);
}

void Ctrl_AddTask(Ctrl *p_ctrl, Ctrl_TaskType type, HitTile *p_tile) {
    // printf("--> add task %d\n", type); fflush(stdout);
    Ctrl_Task task = CTRL_TASK_NULL;
    task.task_type = type;

    task.p_tile = p_tile;

    Ctrl_TaskQueue_Push(p_ctrl->p_task_queue, task);
    // printf("--> added task %d\n", type); fflush(stdout);
}

void Ctrl_Sync(Ctrl *p_ctrl) {
    omp_set_lock(p_ctrl->p_lock_first_host);
    omp_set_lock(p_ctrl->p_lock_ctrl);
    omp_unset_lock(p_ctrl->p_lock_first_host);

    omp_unset_lock(p_ctrl->p_lock_ctrl);
    omp_unset_lock(p_ctrl->p_lock_host);

    omp_set_lock(p_ctrl->p_lock_first_ctrl);
    omp_set_lock(p_ctrl->p_lock_host);
    omp_unset_lock(p_ctrl->p_lock_first_ctrl);
}

void Ctrl_SyncIfSync(Ctrl *p_ctrl) {
     if (p_ctrl->policy == CTRL_POLICY_SYNC) {
        Ctrl_Sync(p_ctrl);
    }
}

void Ctrl_LaunchKernel(Ctrl *p_ctrl, Ctrl_Task task) {
    Ctrl_TaskQueue_Push(p_ctrl->p_task_queue, task);
}

void Ctrl_LaunchHostTask(Ctrl *p_ctrl, Ctrl_Task task) {
    Ctrl_TaskQueue_Push(p_ctrl->p_task_queue, task);
    
    Ctrl_SyncIfSync(p_ctrl);
}

void Ctrl_GlobalSync(Ctrl *p_ctrl) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_GLOBALSYNC, NULL);

    Ctrl_Sync(p_ctrl);
}

void Ctrl_AllocInner(Ctrl *p_ctrl, HitTile *p_tile) {
    p_tile->acumCard = p_tile->origAcumCard[0] += pow(MAX_BLOCK_SIZE, hit_shapeDims(p_tile->shape)) - (p_tile->origAcumCard[0] % (int)pow(MAX_BLOCK_SIZE, hit_shapeDims(p_tile->shape)));
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_ALLOCTILE, p_tile);

    Ctrl_Sync(p_ctrl);
}

void Ctrl_FreeInner(Ctrl *p_ctrl, HitTile *p_tile) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_FREETILE, p_tile);

    Ctrl_Sync(p_ctrl);
}

void Ctrl_MoveToInner(Ctrl *p_ctrl, HitTile *p_tile) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_MOVETO, p_tile);

    Ctrl_SyncIfSync(p_ctrl);
}

void Ctrl_MoveFromInner(Ctrl *p_ctrl, HitTile *p_tile) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_MOVEFROM, p_tile);

    Ctrl_SyncIfSync(p_ctrl);
}

void Ctrl_WaitTileInner(Ctrl *p_ctrl, HitTile *p_tile) {
    Ctrl_AddTask(p_ctrl, CTRL_TASK_TYPE_WAITTILE, p_tile);

    Ctrl_Sync(p_ctrl);
}

/*
 * Function to choose the best implementation alternative
 * Masks: 4bits for Xhi implementations, 5 bits for GPU implementations,
 * 4 bits for CPU implementations, and 1 bit for a Generic implementation
 */
int Ctrl_Dev(Ctrl_Type type, int sum_impl) {
    int result;
    int tmp;
    int pos;

    if (type == CTRL_TYPE_CUDA) {
        tmp = sum_impl & 0b100000000111100000;
        pos = tmp ^ (tmp & (tmp - 1));
        result = pos;
    } else if (type == CTRL_TYPE_CPU) {
        tmp = sum_impl & 0b100000000000011110;
        pos = tmp ^ (tmp & (tmp - 1));
        result = pos;
    } else if (type == CTRL_TYPE_XPHI) {
        tmp = sum_impl & 0b100001111000000000;
        pos = tmp ^ (tmp & (tmp - 1));
        result = pos;
    } else if (type == CTRL_TYPE_OPENCL_GPU) {
        tmp = sum_impl & 0b111110000000000000;
        pos = tmp ^ (tmp & (tmp - 1));
        result = pos;
    } else if (type == CTRL_TYPE_FPGA) {
        tmp = sum_impl & 0b111110000000000000;
        pos = tmp ^ (tmp & (tmp - 1));
        //result = pos;
        result = 1048576;
    } //TODO - FPGA
    return result;
}
