#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>

#include "Ctrl.h"

#define SEED 6834723

/*Tiempos*/
double main_clock;
double exec_clock;

#define THRESH_HOLD_Y 20
#define THRESH_HOLD_U 1
#define THRESH_HOLD_V 1

typedef unsigned char BYTE;


#define N_IMG 3
#define IMG_Y 0
#define IMG_U 1
#define IMG_V 2

/*Tipo de datos*/
Ctrl_NewType(BYTE);

CTRL_KERNEL(Sobel_Operation, GENERIC, KHitTile_BYTE Output, KHitTile_BYTE Input,
        int Width, int Height, int Thresh, int Total_Width, int Total_Height,
{
    // Variable for Gradient in X and Y direction and Final one
    float Gradient_h;
    float Gradient_v;

    // Pixel value's
    char Pixel_Value = 0;

    // Calculating index id
    unsigned int Col_Index = thread_id.y;
    unsigned int Row_Index = thread_id.x;
    if ((Row_Index < Height) && (Col_Index < Width)) {
        if ((Row_Index != 0) && (Col_Index != 0) && 
                (Row_Index != Total_Height - 1) && (Col_Index != Total_Width - 1)
        ) {
            Gradient_v = 
                    -(-hit(Input, (Row_Index - 1) * Width + (Col_Index - 1)) +
                    hit(Input, (Row_Index - 1) * Width + (Col_Index + 1)) -
                    2 * hit(Input, Row_Index * Width + (Col_Index - 1)) +
                    2 * hit(Input, Row_Index * Width + (Col_Index + 1)) -
                    hit(Input, (Row_Index + 1) * Width + (Col_Index - 1)) +
                    hit(Input, (Row_Index + 1) * Width + (Col_Index + 1)));

            Gradient_h = 
                    -(-hit(Input, (Row_Index - 1) * Width + (Col_Index - 1)) -
                    2 * hit(Input, (Row_Index - 1) * Width + Col_Index) -
                    hit(Input, (Row_Index - 1) * Width + (Col_Index + 1)) +
                    hit(Input, (Row_Index + 1) * Width + (Col_Index - 1)) +
                    2 * hit(Input, (Row_Index + 1) * Width + Col_Index) +
                    hit(Input, (Row_Index + 1) * Width + (Col_Index + 1)));

            // Assign to image
            Pixel_Value = (char)sqrt(Gradient_h * Gradient_h +
                                    Gradient_v * Gradient_v);
            if (Pixel_Value >= Thresh)
                hit(Output, (Row_Index - 1) * Width + (Col_Index - 1)) = (char)(255);
            else
                hit(Output, (Row_Index - 1) * Width + (Col_Index - 1)) = (char)(0);
        }
    }
});

CTRL_KERNEL_PROTO(Sobel_Operation, 1, GENERIC, 7, 
        OUT, HitTile_BYTE, Output, 
        IN, HitTile_BYTE, Input, 
        INVAL, int, Width,
        INVAL, int, Height,
        INVAL, int, Thresh,
        INVAL, int, Total_Width,
        INVAL, int, Total_Height
);

void Save_Frame(BYTE *Output_Img[N_IMG], FILE *File_writer, size_t sizes[N_IMG]) {
    for (int i = 0; i < N_IMG; i++) {
        fwrite(Output_Img[i], sizeof(BYTE), sizes[i], File_writer);
    }
}


CTRL_HOST_TASK(Load_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, FILE *File_reader, size_t *sizes,
{
    fread(&(hit(Image_Y, 0)), sizeof(BYTE), sizes[IMG_Y], File_reader);
    fread(&(hit(Image_U, 0)), sizeof(BYTE), sizes[IMG_U], File_reader);
    fread(&(hit(Image_V, 0)), sizeof(BYTE), sizes[IMG_V], File_reader);
});

CTRL_HOST_TASK_PROTO(Load_Frame, 5, 
        OUT, HitTile_BYTE, Image_Y,
        OUT, HitTile_BYTE, Image_U,
        OUT, HitTile_BYTE, Image_V,
        INVAL, FILE *, File_reader,
        INVAL, size_t *, sizes
);

CTRL_HOST_TASK(Put_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, BYTE **buffer_write, size_t *sizes,
{
    memcpy(buffer_write[IMG_Y], &(hit(Image_Y, 0)), sizeof(BYTE) * sizes[IMG_Y]);
    memcpy(buffer_write[IMG_U], &(hit(Image_U, 0)), sizeof(BYTE) * sizes[IMG_U]);
    memcpy(buffer_write[IMG_V], &(hit(Image_V, 0)), sizeof(BYTE) * sizes[IMG_V]);

});

CTRL_HOST_TASK_PROTO(Put_Frame, 5,
        IN, HitTile_BYTE, Image_Y,
        IN, HitTile_BYTE, Image_U,
        IN, HitTile_BYTE, Image_V,
        INVAL, BYTE **, buffer_write,
        INVAL, size_t *, sizes
);

int main(int argc, char *argv[])
{
    main_clock = omp_get_wtime();

    if (argc != 8) {
        printf("Usage: %s <width> <height> <num_frames> <input_yuv_file> "
               "<output_yuv_file> <device> <policy>",
               argv[0]);
        exit(EXIT_FAILURE);
    }

    int Width[3];
    Width[0] = atoi(argv[1]);
    Width[1] = Width[2] = Width[0] / 2; 
    
    int Height[3];
    Height[0] = atoi(argv[2]);
    Height[1] = Height[2] = Height[0] / 2;
    
    int Num_Frames = atoi(argv[3]);
    
    char *Input_Filename = argv[4];
    char *Output_Filename = argv[5];
    int DEVICE = atoi(argv[6]);
    Ctrl_Policy POLICY = (Ctrl_Policy)atoi(argv[7]);

    int Frame_num = 0; // loop variable

    int thresh_hold[N_IMG] = {
            THRESH_HOLD_Y,
            THRESH_HOLD_U,
            THRESH_HOLD_V
    };

    size_t *sizes = (size_t *)malloc(sizeof(size_t) * N_IMG);
    sizes[IMG_Y] = (size_t)(Width[IMG_Y] * Height[IMG_Y]);
    sizes[IMG_U] = (size_t)(Width[IMG_U] * Height[IMG_U]);
    sizes[IMG_V] = (size_t)(Width[IMG_V] * Height[IMG_V]);

    BYTE ***buffer_write;
    
    buffer_write = (BYTE ***)malloc(sizeof(BYTE **) * Num_Frames);

    for (int i = 0; i < Num_Frames; i++) {
        // buffer_read[i] = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);
        buffer_write[i] = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);
        for (int j = 0; j < N_IMG; j++) {
            cudaMallocHost((void **)(&(buffer_write[i][j])), sizes[j] * sizeof(BYTE));
        }
    }

    // File pointer for reading and writting
    FILE *File_writer, *File_reader;

    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("{\"width\":%d,\"height\":%d,\"num_frames\":%d,\"version\":\"ctrl\",\"arch\":\"cuda\",\"device\":%d,", 
				Width[0], Height[0], Num_Frames, DEVICE);
		#ifdef _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
			printf("\"policy\":\"async\",");
		#else  // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
			printf("\"policy\":\"sync\",");
		#endif // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
		fflush(stdout);
	#endif

    #ifdef _CTRL_EXAMPLES_CUDA_INFO_
		#ifndef _CTRL_EXAMPLES_EXP_MODE_
            printf("\n ----------------------- ARGS ----------------------- \n");
            printf("\n WIDTH: %d", Width[0]);
            printf("\n HEIGHT: %d", Height[0]);
            printf("\n NUM_FRAMES: %d", Num_Frames);
            printf("\n DEVICE: %d", DEVICE);
            #ifdef _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
                printf("\n POLICY ASYNC");
            #else  // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
                printf("\n POLICY SYNC");
            #endif // _CTRL_EXAMPLES_CUDA_POLICY_ASYNC_
            printf("\n\n ---------------------------------------------------- \n");
            fflush(stdout);
		#endif // _CTRL_EXAMPLES_EXP_MODE_
    #endif // _CTRL_EXAMPLES_CUDA_INFO_

    Ctrl_Thread threads[N_IMG];
    for (int i = 0; i < N_IMG; i++) {
        Ctrl_ThreadInit(threads[i], Height[i], Width[i]);
    }
    
    if (!(File_reader = fopen(Input_Filename, "rb"))) {
        printf("\nError in opening input file: %s\n", Input_Filename);
        exit(EXIT_FAILURE);
    }
    if (!(File_writer = fopen(Output_Filename, "wb+"))) {
        printf("\nError in opening output file: %s\n", Output_Filename);
        exit(EXIT_FAILURE);
    }

    Ctrl_Thread block;
	Ctrl_ThreadInit(block, BLOCKSIZE_0, BLOCKSIZE_1);
    __ctrl_block__(1)
    {
        PCtrl ctrl = Ctrl_Create(CTRL_TYPE_CUDA, POLICY, DEVICE);

        HitTile_BYTE Input_Img[N_IMG];
        Input_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Input_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Input_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));

        HitTile_BYTE Output_Img[N_IMG];
        Output_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Output_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Output_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));
        

        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime();

        Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                Input_Img[IMG_V], File_reader, sizes
        );
        
        for (Frame_num = 0; Frame_num < Num_Frames; Frame_num++) {
            // Ctrl_MoveTo(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V]);

            for (int i = 0; i < N_IMG; i++) {
                // Ctrl_MoveTo(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V]);
                // Ctrl_MoveTo(ctrl, Input_Img[i]);
                Ctrl_Launch(ctrl, Sobel_Operation, threads[i], block, Output_Img[i], Input_Img[i], 
                        Width[i], Height[i], thresh_hold[i], Width[0], Height[0]
                );
                // Ctrl_MoveFrom(ctrl, Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]);
                // Ctrl_MoveFrom(ctrl, Output_Img[i]);
            }
            if (Frame_num + 1 < Num_Frames) {
                // int frame_aux = Frame_num + 1;
                Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                        Input_Img[IMG_V], File_reader, sizes
                );
            }
            // Ctrl_MoveFrom(ctrl, Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]);
            Ctrl_HostTask(ctrl, Put_Frame, Output_Img[IMG_Y], Output_Img[IMG_U], 
                    Output_Img[IMG_V], buffer_write[Frame_num], sizes
            );
        }

        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime() - exec_clock;

        for (int i = 0; i < Num_Frames; i++) {
            Save_Frame(buffer_write[i], File_writer, sizes);
        }

        Ctrl_Free(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V],
                Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]
        );

        Ctrl_Destroy(ctrl);
    }

    free(sizes);
    fclose(File_reader);
    fclose(File_writer);

    for (int i = 0; i < Num_Frames; i++) {
        for (int j = 0; j < N_IMG; j++) {
            cudaFreeHost(buffer_write[i][j]);
        }
        free(buffer_write[i]);
    }
    free(buffer_write);

    main_clock = omp_get_wtime() - main_clock;
    
    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"main_clock\":%lf,\"exec_clock\":%lf}\n", main_clock, exec_clock);
		fflush(stdout);
	#else
		printf("\n ----------------------- TIME ----------------------- \n\n");
		printf(" Clock main: %lf\n", main_clock);
		printf(" Clock exec: %lf\n", exec_clock);
        printf("\n ---------------------------------------------------- \n");
	#endif

    return 0;
}
