/*
 * <license>
 * 
 * Controller v2.1
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
// #include <sys/mman.h>
#include <cuda.h>
#include <cuda_runtime_api.h>
// #include <cuda_runtime.h>
// #include <helper_cuda.h>

#include "Ctrl.h"

#define SEED 6834723

/*Tiempos*/
double main_clock;
double exec_clock;

#define THRESH_HOLD_Y 20
#define THRESH_HOLD_U 1
#define THRESH_HOLD_V 1

typedef unsigned char BYTE;

#define N_IMG 3
#define IMG_Y 0
#define IMG_U 1
#define IMG_V 2

/*Tipo de datos*/
Ctrl_NewType(BYTE);

CTRL_KERNEL(Sobel_Operation, GENERIC, KHitTile_BYTE Output, KHitTile_BYTE Input,
            int Width, int Height, int Thresh, int Total_Width, int Total_Height, {
                // Variable for Gradient in X and Y direction and Final one
                float Gradient_h;
                float Gradient_v;

                // Pixel value's
                char Pixel_Value = 0;

                // Calculating index id
                unsigned int Col_Index = thread_id.y;
                unsigned int Row_Index = thread_id.x;
                if ((Row_Index < Height) && (Col_Index < Width)) {
                    if ((Row_Index != 0) && (Col_Index != 0) && 
                            (Row_Index != Total_Height - 1) && (Col_Index != Total_Width - 1)
                    ) {
                        Gradient_v = 
                                -(-hit(Input, (Row_Index - 1) * Width + (Col_Index - 1)) +
                                hit(Input, (Row_Index - 1) * Width + (Col_Index + 1)) -
                                2 * hit(Input, Row_Index * Width + (Col_Index - 1)) +
                                2 * hit(Input, Row_Index * Width + (Col_Index + 1)) -
                                hit(Input, (Row_Index + 1) * Width + (Col_Index - 1)) +
                                hit(Input, (Row_Index + 1) * Width + (Col_Index + 1)));

                        Gradient_h = 
                                -(-hit(Input, (Row_Index - 1) * Width + (Col_Index - 1)) -
                                2 * hit(Input, (Row_Index - 1) * Width + Col_Index) -
                                hit(Input, (Row_Index - 1) * Width + (Col_Index + 1)) +
                                hit(Input, (Row_Index + 1) * Width + (Col_Index - 1)) +
                                2 * hit(Input, (Row_Index + 1) * Width + Col_Index) +
                                hit(Input, (Row_Index + 1) * Width + (Col_Index + 1)));

                        // Assign to image
                        Pixel_Value = (char)sqrt(Gradient_h * Gradient_h +
                                                Gradient_v * Gradient_v);
                        if (Pixel_Value >= Thresh)
                            hit(Output, (Row_Index - 1) * Width + (Col_Index - 1)) = (char)(255);
                        else
                            hit(Output, (Row_Index - 1) * Width + (Col_Index - 1)) = (char)(0);
                    }
                }
            });

CTRL_KERNEL_PROTO(Sobel_Operation, 1, GENERIC, 7, 
        OUT, HitTile_BYTE, Output, 
        IN, HitTile_BYTE, Input, 
        INVAL, int, Width,
        INVAL, int, Height,
        INVAL, int, Thresh,
        INVAL, int, Total_Width,
        INVAL, int, Total_Height
);


void Save_Frame(BYTE *Output_Img[N_IMG], FILE *File_writer, size_t sizes[N_IMG]) {
    for (int i = 0; i < N_IMG; i++) {
        fwrite(Output_Img[i], sizeof(BYTE), sizes[i], File_writer);
    }
}

// void Get_Frame(BYTE **buffer_read, BYTE *Input_Img[N_IMG], size_t *sizes) {
//     for (int i = 0; i < N_IMG; i++) {
//         memcpy(Input_Img[i], buffer_read[i], sizeof(BYTE) * sizes[i]);
//     }
// }

// CTRL_HOST_TASK(Get_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
//         HitTile_BYTE Image_V, BYTE **buffer_read, size_t *sizes,
// {
//     memcpy(&(hit(Image_Y, 0)), buffer_read[IMG_Y], sizeof(BYTE) * sizes[IMG_Y]);
//     memcpy(&(hit(Image_U, 0)), buffer_read[IMG_U], sizeof(BYTE) * sizes[IMG_U]);
//     memcpy(&(hit(Image_V, 0)), buffer_read[IMG_V], sizeof(BYTE) * sizes[IMG_V]);
//     // HitTile_BYTE Input_Img[N_IMG] = {Image_Y, Image_U, Image_V};

//     // for (int i = 0; i < N_IMG; i++) {
//     //     fread(&hit(Input_Img[i], 0), sizeof(BYTE), sizes[i], File_reader);
//     // }
//     // size_t read_size;
//     // fread(&(hit(Image_Y, 0)), sizeof(BYTE), sizes[IMG_Y], File_reader);
//     // if (read_size != sizes[IMG_Y]){
//     //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
//     // }
//     // fread(&(hit(Image_U, 0)), sizeof(BYTE), sizes[IMG_U], File_reader);
//     // if (read_size != sizes[IMG_U]){
//     //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
//     // }
//     // fread(&(hit(Image_V, 0)), sizeof(BYTE), sizes[IMG_V], File_reader);
//     // if (read_size != sizes[IMG_V]){
//     //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
//     // }
// /*
//     size_t read_size;
//     if ((read_size = fread(hit(Image_Y, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
//         printf("\nRead operation; Error occured\n");
//         printf("reached end of video file\n");
//         printf("read size is %d\n", read_size);
//         exit(EXIT_FAILURE);
//     }
//     if ((read_size = fread(hit(Image_U, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
//         printf("\nRead operation; Error occured\n");
//         printf("reached end of video file\n");
//         printf("read size is %d\n", read_size);
//         exit(EXIT_FAILURE);
//     }
//     if ((read_size = fread(hit(Image_V, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
//         printf("\nRead operation; Error occured\n");
//         printf("reached end of video file\n");
//         printf("read size is %d\n", read_size);
//         exit(EXIT_FAILURE);
//     }
//     */
// });

// CTRL_HOST_TASK_PROTO(Get_Frame, 5, 
//         OUT, HitTile_BYTE, Image_Y,
//         OUT, HitTile_BYTE, Image_U,
//         OUT, HitTile_BYTE, Image_V,
//         INVAL, BYTE **, buffer_read,
//         INVAL, size_t *, sizes
// );

CTRL_HOST_TASK(Load_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, FILE *File_reader, size_t *sizes,
{
    // HitTile_BYTE Input_Img[N_IMG] = {Image_Y, Image_U, Image_V};

    // for (int i = 0; i < N_IMG; i++) {
    //     fread(&hit(Input_Img[i], 0), sizeof(BYTE), sizes[i], File_reader);
    // }
    // size_t read_size;
    fread(&(hit(Image_Y, 0)), sizeof(BYTE), sizes[IMG_Y], File_reader);
    // if (read_size != sizes[IMG_Y]){
    //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // }
    fread(&(hit(Image_U, 0)), sizeof(BYTE), sizes[IMG_U], File_reader);
    // if (read_size != sizes[IMG_U]){
    //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // }
    fread(&(hit(Image_V, 0)), sizeof(BYTE), sizes[IMG_V], File_reader);
    // if (read_size != sizes[IMG_V]){
    //     printf("error de lectura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // }
/*
    size_t read_size;
    if ((read_size = fread(hit(Image_Y, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
        printf("\nRead operation; Error occured\n");
        printf("reached end of video file\n");
        printf("read size is %d\n", read_size);
        exit(EXIT_FAILURE);
    }
    if ((read_size = fread(hit(Image_U, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
        printf("\nRead operation; Error occured\n");
        printf("reached end of video file\n");
        printf("read size is %d\n", read_size);
        exit(EXIT_FAILURE);
    }
    if ((read_size = fread(hit(Image_V, 0), sizeof(BYTE), sizes[i], File_reader)) != sizes[i]) {
        printf("\nRead operation; Error occured\n");
        printf("reached end of video file\n");
        printf("read size is %d\n", read_size);
        exit(EXIT_FAILURE);
    }
    */
});

CTRL_HOST_TASK_PROTO(Load_Frame, 5, 
        OUT, HitTile_BYTE, Image_Y,
        OUT, HitTile_BYTE, Image_U,
        OUT, HitTile_BYTE, Image_V,
        INVAL, FILE *, File_reader,
        INVAL, size_t *, sizes
);

CTRL_HOST_TASK(Put_Frame, HitTile_BYTE Image_Y, HitTile_BYTE Image_U,
        HitTile_BYTE Image_V, BYTE **buffer_write, size_t *sizes,
{
    memcpy(buffer_write[IMG_Y], &(hit(Image_Y, 0)), sizeof(BYTE) * sizes[IMG_Y]);
    memcpy(buffer_write[IMG_U], &(hit(Image_U, 0)), sizeof(BYTE) * sizes[IMG_U]);
    memcpy(buffer_write[IMG_V], &(hit(Image_V, 0)), sizeof(BYTE) * sizes[IMG_V]);

    // // size_t write_size;
    // fwrite(&(hit(Image_Y, 0)), sizeof(BYTE), sizes[IMG_Y], File_writer);
    // // if (write_size != sizes[IMG_Y]) {
    // //     printf("error de escritura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // // }
    // fwrite(&(hit(Image_U, 0)), sizeof(BYTE), sizes[IMG_U], File_writer);
    // // if (write_size != sizes[IMG_U]) {
    // //     printf("error de escritura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // // }
    // fwrite(&(hit(Image_V, 0)), sizeof(BYTE), sizes[IMG_V], File_writer);
    // if (write_size != sizes[IMG_V]) {
    //     printf("error de escritura\n"); fflush(stdout); exit(EXIT_FAILURE);
    // }
    /*
    HitTile_BYTE Output_Img[N_IMG] = {Image_Y, Image_U, Image_V};

    for (int i = 0; i < N_IMG; i++) {
        fwrite(&hit(Output_Img[i], 0), sizeof(BYTE), sizes[i], File_writer);
    }
    */
});

CTRL_HOST_TASK_PROTO(Put_Frame, 5,
        IN, HitTile_BYTE, Image_Y,
        IN, HitTile_BYTE, Image_U,
        IN, HitTile_BYTE, Image_V,
        INVAL, BYTE **, buffer_write,
        INVAL, size_t *, sizes
);

int main(int argc, char *argv[]) {
    main_clock = omp_get_wtime();

    if (argc != 9) {
        printf("Usage: %s <width> <height> <num_frames> <input_yuv_file> "
               "<output_yuv_file> <device> <platform> <policy>\n",
               argv[0]);
        exit(EXIT_FAILURE);
    }

    int Width[3];
    Width[0] = atoi(argv[1]);
    Width[1] = Width[2] = Width[0] / 2; 
    
    int Height[3];
    Height[0] = atoi(argv[2]);
    Height[1] = Height[2] = Height[0] / 2;
    
    int Num_Frames = atoi(argv[3]);
    
    char *Input_Filename = argv[4];
    char *Output_Filename = argv[5];
    int DEVICE = atoi(argv[6]);
    int PLATFORM = atoi(argv[7]);
    Ctrl_Policy POLICY = (Ctrl_Policy)atoi(argv[8]);

    int Frame_num = 0; // loop variable

    int thresh_hold[N_IMG] = {
            THRESH_HOLD_Y,
            THRESH_HOLD_U,
            THRESH_HOLD_V
    };

    size_t *sizes = (size_t *)malloc(sizeof(size_t) * N_IMG);
    sizes[IMG_Y] = (size_t)(Width[IMG_Y] * Height[IMG_Y]);
    sizes[IMG_U] = (size_t)(Width[IMG_U] * Height[IMG_U]);
    sizes[IMG_V] = (size_t)(Width[IMG_V] * Height[IMG_V]);

    // BYTE ***buffer_read;
    BYTE ***buffer_write;
    
    // BYTE **buffer;
    // buffer = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);
    // for (int i = 0; i < N_IMG; i++) {
    //     buffer[i] = (BYTE *)malloc(sizeof(BYTE) * sizes[i]);
    // }
    // buffer_read = (BYTE ***)malloc(sizeof(BYTE **) * Num_Frames);
    buffer_write = (BYTE ***)malloc(sizeof(BYTE **) * Num_Frames);
    // cudaMallocHost((void **)(&(buffer_read)), Num_Frames * sizeof(BYTE **));
    // cudaMallocHost((void **)(&(buffer_write)), Num_Frames * sizeof(BYTE **));
    // if (buffer_read == NULL) {
    //     printf("error malloc\n"); exit(EXIT_FAILURE); fflush(stdout);
    // }

    for (int i = 0; i < Num_Frames; i++) {
        // buffer_read[i] = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);
        buffer_write[i] = (BYTE **)malloc(sizeof(BYTE *) * N_IMG);
        // cudaMallocHost((void **)(&(buffer_read[i])), N_IMG * sizeof(BYTE *));
        // cudaMallocHost((void **)(&(buffer_write[i])), N_IMG * sizeof(BYTE *));
        // if (buffer_read[i] == NULL) {
        //     printf("error malloc\n"); exit(EXIT_FAILURE); fflush(stdout);
        // }
        for (int j = 0; j < N_IMG; j++) {
            // cudaMallocHost((void **)(&(buffer_read[i][j])), sizes[j] * sizeof(BYTE));
            cudaMallocHost((void **)(&(buffer_write[i][j])), sizes[j] * sizeof(BYTE));
            // buffer_read[i][j] = (BYTE *)malloc(sizeof(BYTE) * sizes[j]);
            // buffer_write[i][j] = (BYTE *)malloc(sizeof(BYTE) * sizes[j]);
            // mlock(buffer_read[i][j], sizes[j] * sizeof(BYTE));
            // mlock(buffer_write[i][j], sizes[j] * sizeof(BYTE));
            // if (buffer_read[i][j] == NULL) {
            //     printf("error malloc\n"); exit(EXIT_FAILURE); fflush(stdout);
            // }
        }
    }

    // File pointer for reading and writting
    FILE *File_writer, *File_reader;


    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("{\"width\":%d,\"height\":%d,\"num_frames\":%d,\"version\":\"ctrl\",\"arch\":\"opencl\",\"device\":%d,\"platform\":%d,", 
				Width[0], Height[0], Num_Frames, DEVICE, PLATFORM);
		#ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\"policy\":\"async\",");
		#else  // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
			printf("\"policy\":\"sync\",");
		#endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
		fflush(stdout);
	#endif

    #ifdef _CTRL_EXAMPLES_OPENCL_GPU_INFO_
		#ifndef _CTRL_EXAMPLES_EXP_MODE_
            printf("\n ----------------------- ARGS ----------------------- \n");
            printf("\n WIDTH: %d", Width[0]);
            printf("\n HEIGHT: %d", Height[0]);
            printf("\n NUM_FRAMES: %d", Num_Frames);
            printf("\n DEVICE: %d", DEVICE);
            printf("\n PLATFORM: %d", PLATFORM);
            #ifdef _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
                printf("\n POLICY ASYNC");
            #else  // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
                printf("\n POLICY SYNC");
            #endif // _CTRL_EXAMPLES_OPENCL_GPU_POLICY_ASYNC_
            printf("\n\n ---------------------------------------------------- \n");
            fflush(stdout);
		#endif // _CTRL_EXAMPLES_EXP_MODE_
    #endif // _CTRL_EXAMPLES_OPENCL_GPU_INFO_

    Ctrl_Thread threads[N_IMG];
    for (int i = 0; i < N_IMG; i++) {
        Ctrl_ThreadInit(threads[i], Height[i], Width[i]);
    }
    
    if (!(File_reader = fopen(Input_Filename, "rb"))) {
        printf("\nError in opening input file: %s\n", Input_Filename);
        exit(EXIT_FAILURE);
    }
    if (!(File_writer = fopen(Output_Filename, "wb+"))) {
        printf("\nError in opening output file: %s\n", Output_Filename);
        exit(EXIT_FAILURE);
    }

    Ctrl_Thread group;
	Ctrl_ThreadInit(group, LOCAL_SIZE_0, LOCAL_SIZE_1);    
    __ctrl_block__(1)
    {
        PCtrl ctrl = Ctrl_Create(CTRL_TYPE_OPENCL_GPU, POLICY, DEVICE, PLATFORM);

        HitTile_BYTE Input_Img[N_IMG];
        Input_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Input_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Input_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));
        //     Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[1], Height[1])),
        //     Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[2], Height[2]))
        // };

        HitTile_BYTE Output_Img[N_IMG];
        Output_Img[IMG_Y] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_Y], Height[IMG_Y]));
        Output_Img[IMG_U] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_U], Height[IMG_U]));
        Output_Img[IMG_V] = Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[IMG_V], Height[IMG_V]));
        
        // HitTile_BYTE Output_Img[N_IMG] = {
        //     Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[0], Height[0])),
        //     Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[1], Height[1])),
        //     Ctrl_Alloc(ctrl, BYTE, hitNewShapeSize(Width[2], Height[2]))
        // };

        Ctrl_Launch(ctrl, Sobel_Operation, threads[0], group, Output_Img[0], Input_Img[0], 
                Width[0], Height[0], thresh_hold[0], Width[0], Height[0]
        );

        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime();

        Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                Input_Img[IMG_V], File_reader, sizes
        );
        
        for (Frame_num = 0; Frame_num < Num_Frames; Frame_num++) {
            // Ctrl_MoveTo(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V]);

            for (int i = 0; i < N_IMG; i++) {
                // Ctrl_MoveTo(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V]);
                Ctrl_MoveTo(ctrl, Input_Img[i]);
                Ctrl_Launch(ctrl, Sobel_Operation, threads[i], group, Output_Img[i], Input_Img[i], 
                        Width[i], Height[i], thresh_hold[i], Width[0], Height[0]
                );
                // Ctrl_MoveFrom(ctrl, Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]);
                Ctrl_MoveFrom(ctrl, Output_Img[i]);
            }
            if (Frame_num + 1 < Num_Frames) {
                // int frame_aux = Frame_num + 1;
                Ctrl_HostTask(ctrl, Load_Frame, Input_Img[IMG_Y], Input_Img[IMG_U], 
                        Input_Img[IMG_V], File_reader, sizes
                );
            }
            // Ctrl_MoveFrom(ctrl, Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]);
            Ctrl_HostTask(ctrl, Put_Frame, Output_Img[IMG_Y], Output_Img[IMG_U], 
                    Output_Img[IMG_V], buffer_write[Frame_num], sizes
            );
        }

        Ctrl_GlobalSync(ctrl);
        exec_clock = omp_get_wtime() - exec_clock;

        for (int i = 0; i < Num_Frames; i++) {
            Save_Frame(buffer_write[i], File_writer, sizes);
        }

        Ctrl_Free(ctrl, Input_Img[IMG_Y], Input_Img[IMG_U], Input_Img[IMG_V],
                Output_Img[IMG_Y], Output_Img[IMG_U], Output_Img[IMG_V]
        );

        Ctrl_Destroy(ctrl);
    }

    // for (int i = 0; i < N_PHASES; i++) {
    // }
    // free(File_reader);
    free(sizes);
    fclose(File_reader);
    fclose(File_writer);

    for (int i = 0; i < Num_Frames; i++) {
        for (int j = 0; j < N_IMG; j++) {
            // cudaFreeHost(buffer_read[i][j]);
            cudaFreeHost(buffer_write[i][j]);
            // munlock(buffer_read[i][j], sizes[j] * sizeof(BYTE));
            // munlock(buffer_write[i][j], sizes[j] * sizeof(BYTE));
            // free(buffer_read[i][j]);
            // free(buffer_write[i][j]);
        }
        // cudaFreeHost(buffer_read[i]);
        // cudaFreeHost(buffer_write[i]);
        // free(buffer_read[i]);
        free(buffer_write[i]);
    }
    // cudaFreeHost(buffer_read);
    // cudaFreeHost(buffer_write);
    // free(buffer_read);
    free(buffer_write);

    main_clock = omp_get_wtime() - main_clock;
    
    #ifdef _CTRL_EXAMPLES_EXP_MODE_
		printf("\"main_clock\":%lf,\"exec_clock\":%lf}\n", main_clock, exec_clock);
		fflush(stdout);
	#else
		printf("\n ----------------------- TIME ----------------------- \n\n");
		printf(" Clock main: %lf\n", main_clock);
		printf(" Clock exec: %lf\n", exec_clock);
        printf("\n ---------------------------------------------------- \n");
	#endif

    return 0;
}
